"""
This is a py.test plugin configuration that initializes a Vagrant VM.

https://pytest.org/latest/writing_plugins.html

Environment variables:

    BURLAP_TEST_BOX
    BURLAP_TEST_PROVIDER
    BURLAP_TEST_REUSE_VM

"""
from __future__ import print_function

from pipes import quote
import logging
import os
import sys

from mock import patch
import pytest

from fabric.api import env, hide, lcd, local, settings
from fabric.state import connections

from burlap.vagrant import vagrant

HERE = os.path.dirname(__file__)
VAGRANT_VERSION = vagrant.version()
MIN_VAGRANT_VERSION = (1, 3)


@pytest.yield_fixture(scope='session', autouse=True)
def setup_package():
    
    # Setup.
    vagrant_box = (os.environ.get('BURLAP_TEST_BOX') or '').strip()
    print('vagrant_box:', vagrant_box)
    if not vagrant_box:
        pytest.skip("Set BURLAP_TEST_BOX to choose a Vagrant base box for functional tests")
    elif vagrant_box == 'localhost':
        # Use no VM. This is intended for use on Travis-CI, where we're already running inside a VM.
        # Be careful with using this option on your development environment, because the tests
        # may disrupt your configuration.
        _configure_logging()
        _target_local_machine()
        yield
    else:
        # Spin up a VM for each test.
        # This is used when running on your development platform.
        # Unfortunately, this isn't supported in Travis-CI, which is already a VM and doesn't support
        # running additional VMs inside of it.
        _check_vagrant_version()
        vagrant_provider = os.environ.get('BURLAP_TEST_PROVIDER')
        reuse_vm = os.environ.get('BURLAP_TEST_REUSE_VM')
        _configure_logging()
        _allow_fabric_to_access_the_real_stdin()
        if not reuse_vm:
            _stop_vagrant_machine()
        _fix_home_directory()
        _init_vagrant_machine(vagrant_box)
#         with settings(warn_only=True):
        _start_vagrant_machine(vagrant_provider)
        _target_vagrant_machine()
        _set_optional_http_proxy()
        #_update_package_index()
        yield
        
        # Teardown.
        if not reuse_vm:
            _stop_vagrant_machine()


def _check_vagrant_version():
    if VAGRANT_VERSION is None:
        pytest.skip("Vagrant is required for functional tests")
    elif VAGRANT_VERSION < MIN_VAGRANT_VERSION:
        pytest.skip("Vagrant >= %s is required for functional tests" % ".".join(map(str, MIN_VAGRANT_VERSION)))


def _configure_logging():
    logger = logging.getLogger('paramiko')
    logger.setLevel(logging.WARN)


def _allow_fabric_to_access_the_real_stdin():
    patcher = patch('fabric.io.sys')
    mock_sys = patcher.start()
    mock_sys.stdin = sys.__stdin__


def _fix_home_directory():
    local("[ `whoami` != `stat -c '%U' ~/.vagrant.d` ] && sudo chown -R `whoami`:`whoami` ~/.vagrant.d || true")


def _init_vagrant_machine(base_box):
    path = os.path.join(HERE, 'Vagrantfile')
    contents = """\
Vagrant.configure(2) do |config|

  config.vm.box = "%s"

  # Speed up downloads using a shared cache across boxes
  if Vagrant.has_plugin?("vagrant-cachier")
    config.cache.scope = :box
  end
  
  config.vm.boot_timeout = 3000

  config.vm.provider "virtualbox" do |vb|
    vb.memory = "2048"
  end

end
""" % base_box
    with open(path, 'w') as vagrantfile:
        vagrantfile.write(contents)


def _start_vagrant_machine(provider):
    print('Starting vagrant with provider %s.' % provider)
    if provider:
        options = ' --provider %s' % quote(provider)
    else:
        options = ''
    with lcd(HERE):
        with settings(warn_only=True):
            ret = local('vagrant up' + options)
            print('ret.return_code:', ret.return_code)
            if ret.return_code:
                # Vagrant is in an inconsistent state, probably because the VM was deleted outside of Vagrant
                # but Vagrant still has the VM's config laying around.
                # So destroy any existing config and re-try.
                _stop_vagrant_machine()
                local('vagrant up' + options)
        #local('export VAGRANT_LOG=DEBUG; vagrant up' + options)


def _stop_vagrant_machine():
    with lcd(HERE):
        with settings(hide('stdout', 'stderr', 'warnings'), warn_only=True):
            local('vagrant halt')
            local('vagrant destroy -f')


def _target_vagrant_machine():
    config = _vagrant_ssh_config()
    print('vagrant.config:', config)
    _set_fabric_env(
        host=config['HostName'],
        port=config['Port'],
        user=config['User'],
        key_filename=config['IdentityFile'].strip('"'),
    )
    _clear_fabric_connection_cache()


def _target_local_machine():
    import getpass
    #http://stackoverflow.com/a/16651742/247542
    _set_fabric_env(
        host='127.0.0.1',
        port='',
        user=getpass.getuser(),
        key_filename=None,
    )
    _clear_fabric_connection_cache()


def _vagrant_ssh_config():
    with lcd(HERE):
        with settings(hide('running')):
            output = local('vagrant ssh-config', capture=True)
    print('output:', output)
    config = {}
    for line in output.splitlines()[1:]:
        key, value = line.strip().split(' ', 2)
        config[key] = value
    return config


def _set_fabric_env(host, port, user, key_filename):
    if port and str(port) != '22':
        env.host_string = env.host = "%s:%s" % (host, port)
    else:
        env.host_string = env.host = host
    env.user = user
    env.key_filename = key_filename
    env.disable_known_hosts = True
    env.abort_on_prompts = True


def _set_optional_http_proxy():
    http_proxy = os.environ.get('BURLAP_HTTP_PROXY')
    if http_proxy is not None:
        env.shell_env['http_proxy'] = http_proxy


def _clear_fabric_connection_cache():
    if env.host_string in connections:
        del connections[env.host_string]


# def _update_package_index():
#     from burlap.system import distrib_family
#     family = distrib_family()
#     if family == 'debian':
#         #from burlap.require.deb import uptodate_index
#         uptodate_index()


@pytest.fixture(scope='session', autouse=True)
def allow_sudo_user(setup_package):
    """
    Fix sudo config if needed

    Some Vagrant boxes come with a too restrictive sudoers config
    and only allow the vagrant user to run commands as root.
    """
    #from burlap.require import file as require_file
    from burlap.files import FileSatchel
    f = FileSatchel()
    f.require(
        '/etc/sudoers.d/burlap',
        contents="vagrant ALL=(ALL) NOPASSWD:ALL\n",
        owner='root',
        mode='440',
        use_sudo=True,
    )



from __future__ import print_function
import os
import sys
# import tempfile
import shutil
import unittest

from burlap import common
# try:
#     import pytest
# except ImportError:
#     pass

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

class ProjectTests(unittest.TestCase):
    
    def setUp(self):
        common.set_verbose(True)
        # Ensure we're in burlap's root directory.
        os.chdir(os.path.abspath(os.path.join(CURRENT_DIR, '../..')))
    
    def test_project(self):
        try:
            project_dir = '/tmp/burlap_test'#tempfile.mkdtemp()
            if not os.path.isdir(project_dir):
                os.makedirs(project_dir)
            bin_dir = os.path.split(sys.executable)[0]
            
            cmd = (
                '. {bin_dir}/activate; '
                'cd {project_dir}; '
                'burlap skel --name=myproject'
            ).format(**locals())
            print(cmd)
            os.system(cmd)
            
            cmd = (
                '. {bin_dir}/activate; '
                'cd {project_dir}; '
                'burlap-admin.py add-role prod dev'
            ).format(**locals())
            print(cmd)
            os.system(cmd)
            
        finally:
            shutil.rmtree(project_dir)
    
    
    def test_find_template(self):
        fn = 'burlap/gitignore.template'
        ret = common.find_template(fn)
        print('ret:', ret)
        assert ret and ret.endswith(fn) 
    
    
    def test_render_to_string(self):
        ret = common.render_to_string(
            'postfix/etc_postfix_sasl_sasl_passwd',
            dict(
                postfix_host='smtp.test.com',
                postfix_port=1234,
                postfix_username='myusername',
                postfix_password='mypassword',
            ))
        print('ret:', ret)
        assert ret == "[smtp.test.com]:1234 myusername:mypassword"



"""
RabbitMQ
============

https://www.rabbitmq.com/

"""
from __future__ import print_function

import sys

from fabric.api import settings

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

DJANGO = 'django'
LOCAL = 'local'

class RabbitMQSatchel(ServiceSatchel):
    
    name = 'rabbitmq'
    
    ## Service options.
    
    ignore_errors = True
    
    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['rabbitmq-server'],
            (UBUNTU, '12.04'): ['rabbitmq-server'],
            (UBUNTU, '14.04'): ['rabbitmq-server'],
            (UBUNTU, '16.04'): ['rabbitmq-server'],
        }
    
    def set_defaults(self):
        self.env.rabbitmq_host = "localhost"
        self.env.vhost = "/"
        self.env.erlang_cookie = ''
        self.env.nodename = "rabbit"
        self.env.user = "guest" # DEPRECATED
        self.env.password = "guest" # DEPRECATED
        self.env.node_ip_address = ''
        self.env.port = 5672
        self.env.erl_args = ""
        self.env.cluster = "no"
        self.env.cluster_config = "/etc/rabbitmq/rabbitmq_cluster.config"
        self.env.logdir = "/var/log/rabbitmq"
        self.env.mnesiadir = "/var/lib/rabbitmq/mnesia"
        self.env.start_args = ""
        self.env.erlang_cookie_template = ''
        self.env.ignore_service_errors = 0
        self.env.management_enabled = False
        self.env.loopback_users = False
        self.env.admin_username = 'admin'
        self.env.user_lookup_method = None # LOCAL|DJANGO
        self.env.users_vhosts = [] # [(user, password, vhost)]
        
        # If true, enables a third-party reposistory to install the most recent version.
        self.env.bleeding = False
            
        self.env.service_commands = {
            START:{
                FEDORA: 'systemctl start rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server start',
            },
            STOP:{
                FEDORA: 'systemctl stop rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server stop',
            },
            DISABLE:{
                FEDORA: 'systemctl disable rabbitmq-server.service',
                UBUNTU: 'chkconfig rabbitmq-server off',
            },
            ENABLE:{
                FEDORA: 'systemctl enable rabbitmq-server.service',
                UBUNTU: 'chkconfig rabbitmq-server on',
            },
            RESTART:{
                FEDORA: 'systemctl restart rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server restart; sleep 5',
            },
            STATUS:{
                FEDORA: 'systemctl status rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server status | cat',
                (UBUNTU, '14.04'): 'service rabbitmq-server status',
                (UBUNTU, '16.04'): 'service rabbitmq-server status | cat',
            },
        }
    
    @task
    def configure_bleeding(self):
        """
        Enables the repository for a most current version on Debian systems.
        
            https://www.rabbitmq.com/install-debian.html
        """
        lm = self.last_manifest
        r = self.local_renderer
        if self.env.bleeding and not lm.bleeding:
            # Install.
            r.append(
                text='deb http://www.rabbitmq.com/debian/ testing main',
                filename='/etc/apt/sources.list.d/rabbitmq.list',
                use_sudo=True)
            r.sudo('cd /tmp; wget -O- https://www.rabbitmq.com/rabbitmq-release-signing-key.asc | sudo apt-key add -')
            r.sudo('apt-get update')
            
        elif not self.env.bleeding and lm.bleeding:
            # Uninstall.
            r.sudo('rm -f /etc/apt/sources.list.d/rabbitmq.list')
            r.sudo('apt-get update')
    
    def render_paths(self):
        r = self.local_renderer
        if self.env.erlang_cookie_template:
            r.env.erlang_cookie = r.format(self.env.erlang_cookie_template)
    
    @task
    def list_vhosts(self):
        """
        Displays a list of configured RabbitMQ vhosts.
        """
        self.sudo('rabbitmqctl list_vhosts')
    
    @task
    def list_users(self):
        """
        Displays a list of configured RabbitMQ users.
        """
        self.sudo('rabbitmqctl list_users')
        
    @task
    def list_queues(self):
        self.sudo('rabbitmqctl list_queues')
        
    @task
    def create_user(self, username, password):
        r = self.local_renderer
        r.env._user = username
        r.env._password = password
        with self.settings(warn_only=True):
            r.sudo('rabbitmqctl add_user {_user} {_password}')

    @task
    def force_stop_and_purge(self):
        """
        Forcibly kills Rabbit and purges all its queues.
        
        For emergency use when the server becomes unresponsive, even to service stop calls.
        
        If this also fails to correct the performance issues, the server may have to be completely
        reinstalled.
        """
        r = self.local_renderer
        r.sudo('killall rabbitmq-server')
        r.sudo('killall beam.smp')
        #TODO:explicitly delete all subfolders, star-delete doesn't work
        r.sudo('rm -Rf /var/lib/rabbitmq/mnesia/*')

    @task
    def add_admin_user(self):
        r = self.local_renderer
        with self.settings(warn_only=True):
            r.sudo('rabbitmqctl add_user {admin_username} {admin_password}')
        r.sudo('rabbitmqctl set_user_tags {admin_username} administrator')
        r.sudo('rabbitmqctl set_permissions -p / {admin_username} ".*" ".*" ".*"')

    @task
    def enable_management_interface(self):
        r = self.local_renderer
        r.sudo('rabbitmq-plugins enable rabbitmq_management')
        r.sudo('service rabbitmq-server restart; sleep 3')
        self.add_admin_user()
        print('You should now be able to access the RabbitMQ web console from:')
        print('\n    http://%s:15672/' % self.genv.host_string)
    
    @task
    def set_loopback_users(self):
        # This allows guest to login through the management interface.
        self.sudo('touch /etc/rabbitmq/rabbitmq.config')
        #self.sudo("echo '[{rabbit, [{loopback_users, []}]}].' >> /etc/rabbitmq/rabbitmq.config")
        self.append(filename='/etc/rabbitmq/rabbitmq.config', text='[{rabbit, [{loopback_users, []}]}].', use_sudo=True)
        with self.settings(warn_only=True):
            self.sudo('service rabbitmq-server restart; sleep 3;')
    
    def get_user_vhosts(self, site=None):
        params = set() # [(user, password, vhost)]
        site = site or ALL
        if self.env.user_lookup_method == DJANGO:
            # Retrieve user settings from one or more assoicated Django sites.
            dj = self.get_satchel('dj')
            for site, site_data in self.iter_sites(site=site, renderer=self.render_paths, no_secure=True):
                if self.verbose:
                    print('!'*80, file=sys.stderr)
                    print('site:', site, file=sys.stderr)
                    
                # Only load site configurations that are allowed for this host.
    #             if target_sites is not None:
    #                 assert isinstance(target_sites, (tuple, list))
    #                 if site not in target_sites:
    #                     continue
                    
                _settings = dj.get_settings(site=site)
                if not _settings:
                    continue
                    
                if hasattr(_settings, 'BROKER_USER') and hasattr(_settings, 'BROKER_VHOST'):
                    if self.verbose:
                        print('RabbitMQ:', _settings.BROKER_USER, _settings.BROKER_VHOST)
                    params.add((_settings.BROKER_USER, _settings.BROKER_PASSWORD, _settings.BROKER_VHOST))
        
        elif self.env.user_lookup_method == LOCAL:
            # Retrieve user settings from our local settings.
            params.update(tuple(_) for _ in self.env.users_vhosts)
        
        elif self.env.user_lookup_method:
            raise NotImplementedError('Unknown user lookup method: %s' % self.env.user_lookup_method)
                
        return params
    
    def _configure(self, site=None, full=0, only_data=0):
        """
        Installs and configures RabbitMQ.
        """

        full = int(full)
        
    #    assert self.env.erlang_cookie
        if full and not only_data:
            packager = self.get_satchel('packager')
            packager.install_required(type=SYSTEM, service=self.name)
        
        #render_paths()
        
#         hostname = get_current_hostname()
#         
#         target_sites = self.genv.available_sites_by_host.get(hostname, None)
        
        r = self.local_renderer
        
        params = self.get_user_vhosts(site=site) # [(user, password, vhost)]

        with settings(warn_only=True):
            self.add_admin_user()
        
        params = sorted(list(params))
        if not only_data:
            for user, password, vhost in params:
                r.env.broker_user = user
                r.env.broker_password = password
                r.env.broker_vhost = vhost
                with settings(warn_only=True):
                    r.sudo('rabbitmqctl add_user {broker_user} {broker_password}')
                    r.sudo('rabbitmqctl add_vhost {broker_vhost}')
                    r.sudo('rabbitmqctl set_permissions -p {broker_vhost} {broker_user} ".*" ".*" ".*"')
                    r.sudo('rabbitmqctl set_permissions -p {broker_vhost} {admin_username} ".*" ".*" ".*"')
                    
        return params    
    
    def record_manifest(self):
        """
        Returns a dictionary representing a serialized state of the service.
        """
        data = super(RabbitMQSatchel, self).record_manifest()
        params = sorted(list(self.get_user_vhosts())) # [(user, password, vhost)]
        data['rabbitmq_all_site_vhosts'] = params
        return data
    
    @task(precursors=['packager', 'user'])
    def configure(self, site=None, **kwargs):
        lm = self.last_manifest

        if self.env.management_enabled != lm.management_enabled:
            self.enable_management_interface()
        
        if self.env.loopback_users != lm.loopback_users:
            self.set_loopback_users()
        
        self.configure_bleeding()
        
        kwargs['site'] = site or ALL
        return self._configure(**kwargs)

rabbitmq = RabbitMQSatchel()



from __future__ import print_function

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

STATIC = 'static'
DYNAMIC = 'dynamic'

class IPSatchel(ServiceSatchel):
    
    name = 'ip'
    
    post_deploy_command = None

    def set_defaults(self):
        self.env.type = DYNAMIC # STATIC
        self.env.interface = 'eth0'
        self.env.address = None
        self.env.network = '192.168.0.0'
        self.env.netmask = '255.255.255.0'
        self.env.broadcast = '10.157.10.255'
        self.env.gateway = '10.157.10.1'
        self.env.dns_nameservers = None
        self.env.interfaces_fn = '/etc/network/interfaces'
        #env.network_restart_command = '/etc/init.d/networking restart'
        self.env.daemon_name = 'networking'
        self.env.service_commands = {
            START:{
                UBUNTU: 'service %s start' % self.env.daemon_name,
            },
            STOP:{
                UBUNTU: 'service %s stop' % self.env.daemon_name,
            },
            DISABLE:{
                UBUNTU: 'chkconfig %s off' % self.env.daemon_name,
            },
            ENABLE:{
                UBUNTU: 'chkconfig %s on' % self.env.daemon_name,
            },
            RESTART:{
                UBUNTU: 'service %s restart' % self.env.daemon_name,
            },
            STATUS:{
                UBUNTU: 'service %s status' % self.env.daemon_name,
            },
        }

    @task
    def static(self):
        """
        Configures the server to use a static IP.
        """
        fn = self.render_to_file('ip/ip_interfaces_static.template')
        r = self.local_renderer
        r.put(local_path=fn, remote_path=r.env.interfaces_fn, use_sudo=True)
    
    @task
    def dynamic(self):
        """
        Configures the server to use a static IP.
        """
        fn = self.render_to_file('ip/ip_interfaces_dynamic.template')
        r = self.local_renderer
        r.put(local_path=fn, remote_path=r.env.interfaces_fn, use_sudo=True)
    
    @task(precursors=['packager', 'user', 'hostname'])
    def configure(self):
        if self.env.type == STATIC:
            self.static()
        elif self.env.type == DYNAMIC:
            self.dynamic()
        else:
            raise NotImplementedError('Unknown type: %s' % self.env.type)
        self.restart()
    
ip = IPSatchel()



from burlap.host import TimezoneSatchel

def test_timezone():    
    ts = TimezoneSatchel()
    ts.verbose = True
    
    current_tz0 = ts.get_current_timezone()
    assert current_tz0 == 'UTC'
    
    ts.env.timezone = 'EST'
    ts.configure()
    
    current_tz = ts.get_current_timezone()
    assert current_tz == 'EST'

    ts.env.timezone = current_tz0
    ts.configure()



# from fabric.api import puts, sudo
# 
# from burlap.require.files import directory as require_directory
# 
# 
# def test_list_partitions():
#     """
#     List disk partitions
#     """
# 
#     from burlap.disk import partitions
# 
#     partitions = partitions()
#     for pname, ptype in partitions.items():
#         puts("%s is %s" % (pname, hex(ptype)))
# 
# 
# def test_format_and_mount():
#     """
#     Format, mount and unmount a block device
#     """
# 
#     from burlap.disk import ismounted, mkfs, mount
# 
#     assert not ismounted('/dev/loop0')
# 
#     try:
#         # Make a loopback block device
#         sudo('dd if=/dev/zero of=bigfile bs=1024 count=30720')
#         sudo('losetup /dev/loop0 bigfile')
# 
#         # Format the block device
#         mkfs('/dev/loop0', 'ext3')
# 
#         # Mount the block device
#         require_directory('/mnt/loop', use_sudo=True)
#         mount('/dev/loop0', '/mnt/loop')
#         assert ismounted('/dev/loop0')
# 
#         # Unmount the block device
#         sudo('umount /dev/loop0')
#         assert not ismounted('/dev/loop0')
# 
#     finally:
#         sudo('umount /dev/loop0', quiet=True)
#         sudo('losetup -d /dev/loop0', quiet=True)
#         sudo('rm -f bigfile', quiet=True)



from __future__ import print_function

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task
from burlap.common import only_hostname

RSYNC = 'rsync'

#DEPRECATED: TODO: remove tarball functionality, and rename to CodeSatchel
class TarballSatchel(Satchel):
    
    name = 'tarball'
    
    def set_defaults(self):
        
        self.env.clean = 1
        
        self.env.gzip = 1
        
        self.env.method = RSYNC
        
        self.env.rsync_source_dir = 'src'
        
        self.env.rsync_source_dirs = [] # This overrides rsync_source_dir
        
        self.env.rsync_target_dir = None
        
        self.env.rsync_target_host = '{user}@{host_string}:'
        
        self.env.rsync_auth = '--rsh "ssh -t -o StrictHostKeyChecking=no -i {key_filename}"'
         
        self.env.rsync_command_template = (
            'rsync '
            '--recursive --verbose --perms --times --links '
            '--compress --copy-links {exclude_str} '
            '--delete --delete-before --force '
            '{rsync_auth} '
            '{rsync_source_dir} '
            '{rsync_target_host}{rsync_target_dir}'
        )
            
        self.env.exclusions = [
            '*_local.py',
            '*.pyc',
            '*.svn',
            '*.tar.gz',
            '*.log',
            'twistd.pid',
            '*.sqlite',
        ]
        
        self.env.dir = '.burlap/tarball_cache'
        
        self.env.extra_dirs = []
        
        self.env.perm_user = 'www-data'
        
        self.env.perm_group = 'www-data'
        
        self.env.perm_chmod = None
        
        self.env.set_permissions = True
    
    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        from burlap.common import get_last_modified_timestamp
        fn = self.env.rsync_source_dir
        if self.verbose:
            print('tarball.fn:', fn)
        data = get_last_modified_timestamp(fn)
        if self.verbose:
            print(data)
        return data
    
    @task
    def set_permissions(self):
        r = self.local_renderer
        if r.env.rsync_target_dir:
            if r.env.perm_chmod:
                r.sudo('chmod -R {perm_chmod} {rsync_target_dir}')
            r.sudo('chown -R {perm_user}:{perm_group} {rsync_target_dir}')
    
    def _run_rsync(self, src, dst):
        print('rsync %s -> %s' % (src, dst))
        r = self.local_renderer
        r.env.hostname = only_hostname(r.genv.host_string)
        real_rsync_target_dir = r.env.rsync_target_dir
        try:
            # Rsync to a temporary directory where we'll have full permissions.
            tmp_dir = '/tmp/tmp_%s_%s' % (self.env.rsync_target_dir.replace('/', '_'), src.replace('/', '_'))
            r.env.rsync_target_dir = tmp_dir
            r.env.rsync_source_dir = src
            r.local(self.env.rsync_command_template)
            
            # Then rsync from the temp directory as sudo to complete the operation.
            r.env.rsync_source_dir = tmp_dir+'/*'
            r.env.rsync_target_dir = real_rsync_target_dir
            r.env.rsync_target_host = ''
            r.env.rsync_auth = ''
            r.sudo(self.env.rsync_command_template)
        finally:
            r.env.rsync_target_dir = real_rsync_target_dir
        
    @task
    def deploy_rsync(self, *args, **kwargs):
        r = self.local_renderer
        
        # Confirm source directories.
        src_dirs = list(self.env.rsync_source_dirs)
        if not src_dirs:
            src_dirs.append(self.env.rsync_source_dir)
            
        # Confirm target directories.
        assert self.env.rsync_target_dir
        
        r.env.exclude_str = ' '.join('--exclude=%s' % _ for _ in self.env.exclusions)
        
        for src_dir in src_dirs:
            self._run_rsync(src=src_dir, dst=self.env.rsync_target_dir)
        
        if self.env.set_permissions:
            self.set_permissions()
    
    @task(precursors=['gitchecker', 'packager', 'apache2', 'pip', 'user'])
    def configure(self, *args, **kwargs):
        if self.env.method == RSYNC:
            self.deploy_rsync(*args, **kwargs)
            
tarball_satchel = TarballSatchel()

deploy = tarball_satchel.configure




from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

class EC2MonitorSatchel(Satchel):
    """
    Manages a script to monitor
    
    http://docs.aws.amazon.com/AmazonCloudWatch/latest/DeveloperGuide/mon-scripts.html
    """
    
    name = 'ec2monitor'
    
    def set_defaults(self):
        self.env.installed = True
        self.env.cron_path = ''
        self.env.install_path = '/home/{user}/aws-scripts-mon'
        self.env.awscreds = 'roles/{role}/aws-{role}.creds'
        self.env.awscreds_install_path = '{install_path}/aws-{role}.creds'
        self.env.options = [
            '--mem-util',
            '--disk-path=/',
            '--disk-space-util',
            '--swap-util',
            # --verify --verbose
        ]
        self.env.access_key_id = None
        self.env.secret_access_key = None
        
        #r = self._get_renderer(verify=False)
        #cmd = #.format(**r.env)
        self.define_cron_job(
            template='etc_crond_ec2monitor',
            script_path='/etc/cron.d/ec2monitor',
            name='default',
        )
        
    def _get_renderer(self, verify=False):
        r = self.local_renderer
        
        r.env.install_path = r.env.install_path.format(**{'user': self.genv.user})
        
        kwargs = dict(
            role=self.genv.ROLE,
            install_path=r.env.install_path,
        )
        r.env.awscreds = r.env.awscreds.format(**kwargs)
        r.env.awscreds_install_path = r.env.awscreds_install_path.format(**kwargs)
        
        options = self.env.options
        if verify:
            options.extend(['--verify --verbose'])
        
        r.env.command_options = ' '.join(options)
        return r
    
    def _get_check_command(self):
        return 'cd {install_path}; export AWS_CREDENTIAL_FILE={awscreds_install_path}; ./mon-put-instance-data.pl {command_options}'
    
    @task
    def verify(self):
        r = self._get_renderer(verify=True)
        r.run(self._get_check_command())
    
    @task
    def check(self):
        r = self._get_renderer(verify=False)
        r.run(self._get_check_command())
    
    @task
    def install(self):
        r = self._get_renderer()
        r.sudo('apt-get install --yes unzip libwww-perl libdatetime-perl')
        r.run('cd ~; curl http://aws-cloudwatch.s3.amazonaws.com/downloads/CloudWatchMonitoringScripts-1.2.1.zip -O')
        r.run('cd ~; unzip -o CloudWatchMonitoringScripts-1.2.1.zip')
        r.run('cd ~; rm CloudWatchMonitoringScripts-1.2.1.zip')
        r.put(
            local_path=self.env.awscreds.format(role=self.genv.ROLE),
            remote_path=r.env.awscreds_install_path,
        )
        self.install_cron_job(
            name='default',
            extra=dict(
                command=self._get_check_command().format(**r.env)
            ))
    
    @task
    def uninstall(self):
        #todo
        pass
    
    @task(precursors=['packager', 'user'])
    def configure(self):
        """
        Executed when your settings have changed since the last deployment.
        Run commands to apply changes here.
        """
        self.install() 
                
ec2monitor = EC2MonitorSatchel()



from __future__ import print_function
from __future__ import absolute_import

import copy
import os
import re
import sys
import types
import importlib
import pkgutil
import inspect
import warnings
from pprint import pprint

try:
    from fabric.api import env
    from fabric.tasks import WrappedCallableTask
    from fabric.utils import _AliasDict
    from fabric.api import hide, settings
    from fabric.decorators import task, runs_once

    import yaml
    
    # Variables cached per-role. Must be after deepcopy.
    env._rc = type(env)()

    # Force dictionaries to be serialized in multi-line format.
    def _represent_dictorder(self, data):
        return self.represent_mapping(u'tag:yaml.org,2002:map', data.items())
    
    def _represent_tuple(self, data):
        return self.represent_sequence(u'tag:yaml.org,2002:seq', data)
    
    def _construct_tuple(self, node):
        return tuple(self.construct_sequence(node))
    
    def _represent_function(self, data):
        return self.represent_scalar(u'tag:yaml.org,2002:null', u'null')
    
    yaml.add_representer(type(env), _represent_dictorder)
    yaml.add_representer(_AliasDict, _represent_dictorder)
    #yaml.add_representer(tuple, _represent_tuple) # we need tuples for hash keys
    yaml.add_constructor(u'tag:yaml.org,2002:python/tuple', _construct_tuple)
    yaml.add_representer(types.FunctionType, _represent_function)

except ImportError as e:
    print('Unable to initialize yaml: %s' % e, file=sys.stderr)

try:
    env
except NameError as e:
    print('Unable to initialize env: %s' % e, file=sys.stderr)
    env = None
    env_default = {}

try:
    from . import common
    Satchel = common.Satchel
    ServiceSatchel = common.ServiceSatchel
    ContainerSatchel = common.ContainerSatchel
    env_default = common.save_env()
except (ImportError, NameError) as e:
    print('Unable to initialize common: %s' % e, file=sys.stderr)
    common = None

try:
    from . import debug
except (ImportError, NameError) as e:
    print('Unable to initialize debug: %s' % e, file=sys.stderr)
    debug = None

VERSION = (0, 9, 9)
__version__ = '.'.join(map(str, VERSION))

burlap_populate_stack = int(os.environ.get('BURLAP_POPULATE_STACK', 1))
no_load = int(os.environ.get('BURLAP_NO_LOAD', 0))

def _get_environ_handler(name, d):
    """
    Dynamically creates a Fabric task for each configuration role.
    """
    
    def func(site=None, **kwargs):
        from fabric import state
        
        # We can't auto-set default_site, because that break tasks that have
        # to operate over multiple sites.
        # If a task requires a site, it can pull from default_site as needed.
        #site = site or d.get('default_site') or env.SITE
        
        BURLAP_SHELL_PREFIX = int(os.environ.get('BURLAP_SHELL_PREFIX', '0'))
        if BURLAP_SHELL_PREFIX:
            print('#!/bin/bash')
            print('# Generated with:')
            print('#')
            print('#     export BURLAP_SHELL_PREFIX=1; export BURLAP_COMMAND_PREFIX=0; fab %s' % (' '.join(sys.argv[1:]),))
            print('#')
            
        BURLAP_COMMAND_PREFIX = int(os.environ.get('BURLAP_COMMAND_PREFIX', '1'))
        with_args = []
        if not BURLAP_COMMAND_PREFIX:
            for k in state.output:
                state.output[k] = False
    
        hostname = kwargs.get('hostname')
        hostname = hostname or kwargs.get('name')
        hostname = hostname or kwargs.get('hn')
        hostname = hostname or kwargs.get('h')

        verbose = int(kwargs.get('verbose', '0'))
        common.set_verbose(verbose)
        
        # Load environment for current role.
        env.update(env_default)
        env[common.ROLE] = os.environ[common.ROLE] = name
        if site:
            env[common.SITE] = os.environ[common.SITE] = site
        env.update(d)
        
        # Load host retriever.
        retriever = None
        if env.hosts_retriever:
            # Dynamically retrieve hosts.
#             module_name = '.'.join(env.hosts_retriever.split('.')[:-1])
#             func_name = env.hosts_retriever.split('.')[-1]
#             retriever = getattr(importlib.import_module(module_name), func_name)
            retriever = common.get_hosts_retriever()
            if verbose:
                print('Using retriever:', env.hosts_retriever, retriever)
        
        # Load host translator.
        translator = None
        if hostname:
            # Filter hosts list by a specific host name.
            module_name = '.'.join(env.hostname_translator.split('.')[:-1])
            func_name = env.hostname_translator.split('.')[-1]
            translator = getattr(importlib.import_module(module_name), func_name)
        
        # Re-load environment for current role, incase loading
        # the retriever/translator reset some environment values.
        env.update(env_default)
        env[common.ROLE] = os.environ[common.ROLE] = name
        if site:
            env[common.SITE] = os.environ[common.SITE] = site
        env.update(d)
        
        # Dynamically retrieve hosts.
        if env.hosts_retriever:
            if verbose:
                print('Building host list with retriever %s...' % env.hosts_retriever)
            env.hosts = list(retriever(site=site))
            if verbose:
                print('Found hosts:')
                print(env.hosts)

        # Filter hosts list by a specific host name.
        if hostname:
            _hostname = hostname
            hostname = translator(hostname=hostname)
            _hosts = env.hosts
            env.hosts = [_ for _ in env.hosts if _ == hostname]
            assert env.hosts, \
                'Hostname %s does not match any known hosts.' % (_hostname,)
                
        if env.is_local is None:
            if env.hosts:
                env.is_local = 'localhost' in env.hosts or '127.0.0.1' in env.hosts
            elif env.host_string:
                env.is_local = 'localhost' in env.host_string or '127.0.0.1' in env.host_string

        for cb in common.post_role_load_callbacks:
            cb()
        
        print('Loaded role %s.' % (name,), file=sys.stderr)
    func.__doc__ = 'Sets enivronment variables for the "%s" role.' % (name,)
    return func

def update_merge(d, u):
    """
    Recursively merges two dictionaries.
    
    Uses fabric's AttributeDict so you can reference values via dot-notation.
    e.g. env.value1.value2.value3...
    
    http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """
    import collections
    for k, v in u.iteritems():
        if isinstance(v, collections.Mapping):
            r = update_merge(d.get(k, dict()), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d

def find_yaml_settings_fn(name, local=False, fn='settings.yaml'):
    if local:
        settings_fn = os.path.join(common.ROLE_DIR, name, 'settings_local.yaml')
    else:
        settings_fn = os.path.join(common.ROLE_DIR, name, fn)
    if os.path.isfile(settings_fn):
        return settings_fn

def load_yaml_settings(name, priors=None, verbose=0):
    verbose = int(verbose)
    config = type(env)()
    if priors is None:
        priors = set()
    if name in priors:
        return config
    priors.add(name)
    
    settings_fn = find_yaml_settings_fn(name)
    if not settings_fn:
        warnings.warn('Warning: Could not find Yaml settings for role %s.' % (name,))
        return config
    if verbose:
        print('Loading settings:', settings_fn)
    config.update(yaml.safe_load(open(settings_fn)) or type(env)())
    
    if 'inherits' in config:
        parent_name = config['inherits']
        del config['inherits']
        parent_config = load_yaml_settings(
            parent_name,
            priors=priors,
            verbose=verbose)
        parent_config.update(config)
        config = parent_config
    
    # Load includes.
    includes = config.pop('includes', [])
    load_includes = []
    for _include_fn in includes:
        load_includes.append(_include_fn)
        include_fn = find_yaml_settings_fn(name=name, fn=_include_fn)
        assert include_fn, 'Invalid include file: %s' % _include_fn
        if verbose:
            print('Loading include settings:', include_fn)
        data = yaml.safe_load(open(include_fn))
        config.update(data)
    
    # Load local overrides.
    settings_local_fn = find_yaml_settings_fn(name, local=True)
    if settings_local_fn:
        if verbose:
            print('Loading local settings:', settings_local_fn)
        data = yaml.safe_load(open(settings_local_fn)) or type(env)()
        includes = data.pop('includes', [])
        config.update(data)
        
        # Load local includes.
        for _include_fn in includes:
            load_includes.append(_include_fn)
            include_fn = find_yaml_settings_fn(name=name, fn=_include_fn)
            assert include_fn, 'Invalid include file: %s' % _include_fn
            if verbose:
                print('Loading include settings:', include_fn)
            data = yaml.safe_load(open(include_fn))
            config.update(data)
    
    config['includes'] = load_includes
    
    return config

try:
    @task
    @runs_once
    def shell(*args, **kwargs):
        return debug.debug.shell(*args, **kwargs)
except NameError:
    pass

def populate_fabfile():
    """
    Automatically includes all submodules and role selectors
    in the top-level fabfile using spooky-scary black magic.
    
    This allows us to avoid manually declaring imports for every module, e.g.
    
        import burlap.pip
        import burlap.vm
        import burlap...
    
    which has the added benefit of allowing us to manually call the commands
    without typing "burlap".
    
    This is soley for convenience. If not needed, it can be disabled
    by specifying the environment variable:
    
        export BURLAP_POPULATE_STACK=0
    """
    stack = inspect.stack()
    fab_frame = None
    for frame_obj, script_fn, line, _, _, _ in stack:
        if 'fabfile.py' in script_fn:
            fab_frame = frame_obj
            break
    if not fab_frame:
        return
    try:
        locals_ = fab_frame.f_locals
        for module_name, module in sub_modules.iteritems():
            locals_[module_name] = module
        for role_name, role_func in role_commands.iteritems():
            assert role_name not in sub_modules, \
                ('The role %s conflicts with a built-in submodule. '
                 'Please choose a different name.') % (role_name)
            locals_[role_name] = role_func
        locals_['common'] = common
        
        # Put all debug commands into the global namespace.
        
#         for _debug_name in debug.debug.get_tasks():
#             print('_debug_name:', _debug_name)

        locals_['shell'] = shell#debug.debug.shell
        
        # Put all virtual satchels in the global namespace so Fabric can find them.
        for _module_alias in common._post_import_modules:
            exec("import %s" % _module_alias) # pylint: disable=exec-used
            locals_[_module_alias] = locals()[_module_alias]

    finally:
        del stack

def load_role_handler(name):
    _config = load_yaml_settings(name)
    _f = _get_environ_handler(name, _config)
    _f = WrappedCallableTask(_f, name=name)
    return _f

# Dynamically create a Fabric task for each role.
role_commands = {}
if common and not no_load:
    if os.path.isdir(common.ROLE_DIR):
        for _name in os.listdir(common.ROLE_DIR):
            _settings_fn = os.path.join(common.ROLE_DIR, _name, 'settings.yaml')
            if _name.startswith('.') or not os.path.isfile(_settings_fn):
                continue
            _f = load_role_handler(_name)
            _var_name = 'role_' + _name
            _cmd = "%s = _f" % (_var_name,)
            exec(_cmd) # pylint: disable=exec-used
            role_commands[_var_name] = _f

    # Auto-import all sub-modules.
    sub_modules = {}
    sub_modules['common'] = common
    __all__ = []
    for loader, module_name, is_pkg in  pkgutil.walk_packages(__path__):
        if module_name in locals():
            continue
        if module_name.startswith('tests'):
            continue
        __all__.append(module_name)
#         print('Importing: %s' % module_name, file=sys.stderr)
        module = loader.find_module(module_name).load_module(module_name)
        sub_modules[module_name] = module
    
    if burlap_populate_stack:
        populate_fabfile()
    
    # Execute any callbacks registered by sub-modules.
    # These are useful for calling inter-sub-module functions
    # after the modules tasks are registered so task names don't get
    # mistakenly registered under the wrong module.
    for cb in common.post_callbacks:
        cb()



"""
Helper functions for sending a notification email after each deployment.
"""
from __future__ import print_function

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

class DeploymentNotifierSatchel(Satchel):
    
    name = 'deploymentnotifier'
    
    def set_defaults(self):
    
        self.env.email_enabled = False
        self.env.email_host = None
        self.env.email_port = 587
        self.env.email_host_user = None
        self.env.email_host_password = None
        self.env.email_use_tls = True
        self.env.email_recipient_list = []

    def send_email(self, subject, message, from_email=None, recipient_list=None):
        import smtplib
        from email.mime.text import MIMEText
        
        recipient_list = recipient_list or []
        if not recipient_list:
            return
        
        from_email = from_email or self.env.email_host_user
        
        msg = MIMEText(message)
        
        # me == the sender's email address
        # you == the recipient's email address
        msg['Subject'] = subject
        msg['From'] = from_email
        msg['To'] = '; '.join(recipient_list)
        
        # Send the message via our own SMTP server, but don't include the
        # envelope header.
        #if self.verbose:
        print('Attempting to send mail using %s@%s...' \
            % (self.env.email_host_user, self.env.email_host))
        s = smtplib.SMTP(self.env.email_host, self.env.email_port)
        s.ehlo()
        s.starttls()
        s.ehlo()
        s.login(self.env.email_host_user, self.env.email_host_password)
        s.sendmail(from_email, recipient_list, msg.as_string())
        s.quit()

    @task
    def notify_post_deployment(self, subject=None, message=None, force=0):
        from burlap import common
        force = int(force)
        subject = subject or '%s Deployment Complete' % self.genv.ROLE.title()
        message = message or 'Deployment to %s is complete.' % self.genv.ROLE
        if self.dryrun:
            self.print_command('echo -e "{body}" | mail -s "{subject}" {recipients}'.format(
                recipients=','.join(self.env.email_recipient_list),
                body=message.replace('\n', '\\n'),
                subject=subject,
            ))
        elif force or (self.env.email_enabled and self.genv.host_string == self.genv.hosts[-1]):
            self.send_email(
                subject=subject,
                message=message,
                recipient_list=self.env.email_recipient_list)

    @task
    def configure(self):
        pass

class LoginNotifierSatchel(Satchel):
    
    name = 'loginnotifier'
    
    def set_defaults(self):
        self.env.sysadmin_email = 'root@localhost'
        self.env.script_template = 'notifier/loginnotifier.template.sh'
        self.env.script_installation_path = '/etc/profile.d/loginnotifier.sh'
        self.env.script_user = 'root'
        self.env.script_group = 'root'
        self.env.script_chmod = 'u+rx,g+rx'
    
    @task(precursors=['packager', 'user'])
    def configure(self):
        r = self.local_renderer
        if self.env.enabled:
            fn = self.render_to_file(self.env.script_template)
            r.put(
                local_path=fn,
                remote_path=self.env.script_installation_path, use_sudo=True)
            r.sudo('chown {script_user}:{script_group} {script_installation_path}')
            r.sudo('chmod {script_chmod} {script_installation_path}')
        else:
            r.sudo('rm {script_installation_path}')


deployment_notifier = DeploymentNotifierSatchel()
notify_post_deployment = deployment_notifier.notify_post_deployment
send_email = deployment_notifier.send_email

login_notifier = LoginNotifierSatchel()



from __future__ import print_function

import re

from fabric.api import hide

from burlap.constants import *
from burlap import ContainerSatchel
from burlap.decorators import task

def _to_int(val):
    try:
        return int(val)
    except ValueError:
        return val

class VagrantSatchel(ContainerSatchel):
    
    name = 'vagrant'
    
    def set_default(self):
        
        self.env.vagrant_box = '?'
        self.env.vagrant_provider = '?'
        self.env.vagrant_shell_command = 'vagrant ssh'

    
    def ssh_config(self, name=''):
        """
        Get the SSH parameters for connecting to a vagrant VM.
        """
        r = self.local_renderer
        with self.settings(hide('running')):
            output = r.local('vagrant ssh-config %s' % name, capture=True)
    
        config = {}
        for line in output.splitlines()[1:]:
            key, value = line.strip().split(' ', 2)
            config[key] = value
        return config


    def _get_settings(self, config):
        settings = {}
    
        user = config['User']
        hostname = config['HostName']
        port = config['Port']
    
        # Build host string
        host_string = "%s@%s:%s" % (user, hostname, port)
    
        settings['user'] = user
        settings['hosts'] = [host_string]
        settings['host_string'] = host_string
    
        # Strip leading and trailing double quotes introduced by vagrant 1.1
        settings['key_filename'] = config['IdentityFile'].strip('"')
    
        settings['forward_agent'] = (config.get('ForwardAgent', 'no') == 'yes')
        settings['disable_known_hosts'] = True
    
        return settings


    @task
    def version(self):
        """
        Get the Vagrant version.
        """
        r = self.local_renderer
        with self.settings(hide('running', 'warnings'), warn_only=True):
            res = r.local('vagrant --version', capture=True)
        if res.failed:
            return None
        line = res.splitlines()[-1]
        version = re.match(r'Vagrant (?:v(?:ersion )?)?(.*)', line).group(1)
        return tuple(_to_int(part) for part in version.split('.'))
    
    
    @task
    def setup(self, name=''):
        r = self.local_renderer
        _settings = self._get_settings(self.ssh_config(name=name))
        if self.verbose:
            print(_settings)
        r.genv.update(_settings)
    
        
    @task
    def init(self):
        r = self.local_renderer
        r.local('vagrant init {box}')
    
        
    @task
    def up(self):
        r = self.local_renderer
        r.local('vagrant up --provider={provider}')
    
        
    @task
    def shell(self):
        r = self.local_renderer
        self.setup()
        r.local(r.env.shell_command)
    
    
    @task
    def destroy(self):
        r = self.local_renderer
        r.local('vagrant destroy')
    
    
    @task
    def upload(self, src, dst=None):
        r = self.local_renderer
        r.put(local_path=src, remote_path=dst)
    
    
    #http://serverfault.com/a/758017/41252
    @task
    def ssh(self):
        r = self.local_renderer
        self.setup()
        hostname, port = r.genv.host_string.split('@')[-1].split(':')
        r.local((
            'ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
            '-i %s %s@%s -p %s') % (r.genv.key_filename, r.genv.user, hostname, port))


    def _settings_dict(self, config):
        settings = {}
    
        user = config['User']
        hostname = config['HostName']
        port = config['Port']
    
        # Build host string
        host_string = "%s@%s:%s" % (user, hostname, port)
    
        settings['user'] = user
        settings['hosts'] = [host_string]
        settings['host_string'] = host_string
    
        # Strip leading and trailing double quotes introduced by vagrant 1.1
        settings['key_filename'] = config['IdentityFile'].strip('"')
    
        settings['forward_agent'] = (config.get('ForwardAgent', 'no') == 'yes')
        settings['disable_known_hosts'] = True
    
        return settings


    @task
    def vagrant(self, name=''):
        """
        Run the following tasks on a vagrant box.
    
        First, you need to import this task in your ``fabfile.py``::
    
            from fabric.api import *
            from burlap.vagrant import vagrant
    
            @task
            def some_task():
                run('echo hello')
    
        Then you can easily run tasks on your current Vagrant box::
    
            $ fab vagrant some_task
    
        """
        r = self.local_renderer
        config = self.ssh_config(name)
    
        extra_args = self._settings_dict(config)
        r.genv.update(extra_args)


    def vagrant_settings(self, name='', *args, **kwargs):
        """
        Context manager that sets a vagrant VM
        as the remote host.
    
        Use this context manager inside a task to run commands
        on your current Vagrant box::
    
            from burlap.vagrant import vagrant_settings
    
            with vagrant_settings():
                run('hostname')
        """
        config = self.ssh_config(name)
    
        extra_args = self._settings_dict(config)
        kwargs.update(extra_args)
    
        return self.settings(*args, **kwargs)
    
    
    def status(self, name='default'):
        """
        Get the status of a vagrant machine
        """
        machine_states = dict(self._status())
        return machine_states[name]
    
    
    def _status(self):
        if self.version() >= (1, 4):
            return self._status_machine_readable()
        else:
            return self._status_human_readable()
    
    
    def _status_machine_readable(self):
        with self.settings(hide('running')):
            output = self.local('vagrant status --machine-readable', capture=True)
        tuples = [tuple(line.split(',')) for line in output.splitlines() if line.strip() != '']
        return [(target, data) for timestamp, target, type_, data in tuples if type_ == 'state-human-short']
    
    
    def _status_human_readable(self):
        with self.settings(hide('running')):
            output = self.local('vagrant status', capture=True)
        lines = output.splitlines()[2:]
        states = []
        for line in lines:
            if line == '':
                break
            target = line[:25].strip()
            state = re.match(r'(.{25}) ([^\(]+)( \(.+\))?$', line).group(2)
            states.append((target, state))
        return states
    
    
    def machines(self):
        """
        Get the list of vagrant machines
        """
        return [name for name, state in self._status()]
    
    
    def base_boxes(self):
        """
        Get the list of vagrant base boxes
        """
        return sorted(list(set([name for name, provider in self._box_list()])))
    
    
    def _box_list(self):
        if self.version() >= (1, 4):
            return self._box_list_machine_readable()
        else:
            return self._box_list_human_readable()
    
    
    def _box_list_machine_readable(self):
        r = self.local_renderer
        with self.settings(hide('running')):
            output = r.local('vagrant box list --machine-readable', capture=True)
        tuples = [tuple(line.split(',')) for line in output.splitlines() if line.strip() != '']
        res = []
        for timestamp, target, type_, data in tuples:
            if type_ == 'box-name':
                box_name = data
            elif type_ == 'box-provider':
                box_provider = data
                res.append((box_name, box_provider))
            else:
                raise ValueError('Unknown item type')
        return res
    
    
    def _box_list_human_readable(self):
        r = self.local_renderer
        with self.settings(hide('running')):
            output = r.local('vagrant box list', capture=True)
        lines = output.splitlines()
        res = []
        for line in lines:
            box_name = line[:25].strip()
            mo = re.match(r'.{25} \((.+)\)$', line)
            box_provider = mo.group(1) if mo is not None else 'virtualbox'
            res.append((box_name, box_provider))
        return res

vagrant = VagrantSatchel()



from __future__ import print_function

import os
import tempfile
from pprint import pprint

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

class PackagerSatchel(Satchel):
    
    name = 'packager'
    
    def set_defaults(self):
        self.env.apt_requirments_fn = 'apt-requirements.txt'
        self.env.yum_requirments_fn = 'yum-requirements.txt'
        self.env.initial_upgrade = True
        self.env.apt_packages = None
        self.env.yum_packages = None
        self.env.do_reboots = True
    
    def record_manifest(self):
        """
        Returns a dictionary representing a serialized state of the service.
        """
        data = {}
        data['required_packages'] = self.install_required(type=SYSTEM, verbose=False, list_only=True)
        data['required_packages'].sort()
        data['custom_packages'] = self.install_custom(list_only=True)
        data['custom_packages'].sort()
        data['repositories'] = self.get_repositories()
        return data
    
    @task
    def prepare(self):
        """
        Preparse the packaging system for installations.
        """
        packager = self.packager
        if packager == APT:
            self.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq update')
        elif packager == YUM:
            self.sudo('yum update')
        else:
            raise Exception('Unknown packager: %s' % (packager,))

    @task
    def install_apt(self, fn=None, package_name=None, update=0, list_only=0):
        """
        Installs system packages listed in apt-requirements.txt.
        """
        r = self.local_renderer
        assert self.genv[ROLE]
        apt_req_fqfn = fn or self.find_template(self.env.apt_requirments_fn)
        if not apt_req_fqfn:
            return []
        assert os.path.isfile(apt_req_fqfn)
        
        lines = list(self.env.apt_packages or [])
        for _ in open(apt_req_fqfn).readlines():
            if _.strip() and not _.strip().startswith('#') \
            and (not package_name or _.strip() == package_name):
                lines.extend(_pkg.strip() for _pkg in _.split(' ') if _pkg.strip()) 
        
        if list_only:
            return lines
        
        tmp_fn = r.write_temp_file('\n'.join(lines))
        apt_req_fqfn = tmp_fn
        
        if not self.genv.is_local:
            r.put(local_path=tmp_fn, remote_path=tmp_fn)
            apt_req_fqfn = self.genv.put_remote_path
        r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq update --fix-missing')
        r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq install `cat "%s" | tr "\\n" " "`' % apt_req_fqfn)

    @task
    def install_yum(self, fn=None, package_name=None, update=0, list_only=0):
        """
        Installs system packages listed in yum-requirements.txt.
        """
        assert self.genv[ROLE]
        yum_req_fn = fn or self.find_template(self.genv.yum_requirments_fn)
        if not yum_req_fn:
            return []
        assert os.path.isfile(yum_req_fn)
        update = int(update)
        if list_only:
            return [
                _.strip() for _ in open(yum_req_fn).readlines()
                if _.strip() and not _.strip.startswith('#')
                and (not package_name or _.strip() == package_name)
            ]
        if update:
            self.sudo_or_dryrun('yum update --assumeyes')
        if package_name:
            self.sudo_or_dryrun('yum install --assumeyes %s' % package_name)
        else:
            if self.genv.is_local:
                self.put_or_dryrun(local_path=yum_req_fn)
                yum_req_fn = self.genv.put_remote_fn
            self.sudo_or_dryrun('yum install --assumeyes $(cat %(yum_req_fn)s)' % yum_req_fn)

    @task
    def install_custom(self, *args, **kwargs):
        """
        Installs all system packages listed in the appropriate
        <packager>-requirements.txt.
        """
        packager = self.packager
        if packager == APT:
            return self.install_apt(*args, **kwargs)
        elif packager == YUM:
            return self.install_yum(*args, **kwargs)
        else:
            raise Exception('Unknown packager: %s' % (packager,))
    
    @task
    def kill_apt_get(self):
        r = self.local_renderer
        r.sudo('killall apt-get')
        r.sudo('DEBIAN_FRONTEND=noninteractive dpkg --configure -a')
    
    @task
    def refresh(self, *args, **kwargs):
        """
        Updates/upgrades all system packages.
        """
        r = self.local_renderer
        packager = self.packager
        if packager == APT:
            r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq update --fix-missing')
        elif packager == YUM:
            raise NotImplementedError
            #return upgrade_yum(*args, **kwargs)
        else:
            raise Exception('Unknown packager: %s' % (packager,))

    @task
    def upgrade(self):
        """
        Updates/upgrades all system packages.
        """
        r = self.local_renderer
        packager = self.packager
        if packager == APT:
            r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq upgrade')
            r.sudo('DEBIAN_FRONTEND=noninteractive apt-get dist-upgrade -yq')
        elif packager == YUM:
            raise NotImplementedError
        else:
            raise Exception('Unknown packager: %s' % (packager,))

    @task
    def get_repositories(self, typ=None, service=None):
        
        service = (service or '').strip().upper()
        typ = (typ or '').lower().strip()
        assert not typ or typ in PACKAGE_TYPES, \
            'Unknown package type: %s' % (typ,)
        
        repositories = {} # {typ: [repos]}
        
        for satchel_name, satchel in self.all_other_enabled_satchels.items():
            
            if service and satchel_name.upper() != service:
                continue
            
            if hasattr(satchel, 'packager_repositories'):
                repos = satchel.packager_repositories
                for repo_type, repo_lst in repos.items():
                    assert isinstance(repo_lst, (tuple, list)), \
                        'Invalid repo list for satchel %s.' % satchel_name
                    for _name in repo_lst:
                        # Can be string (for APT) or tuple of strings (for APT SOURCE)
                        if isinstance(_name, basestring):
                            _name = _name.strip()
                            assert _name, 'Invalid repo name for satchel %s.' % satchel_name
                    repositories.setdefault(repo_type, [])
                    repositories[repo_type].extend(repo_lst)
                    
        for repo_type in repositories:
            repositories[repo_type].sort()
                    
        return repositories
    
    @task
    def install_repositories(self, *args, **kwargs):
        r = self.local_renderer
        repos = self.get_repositories(*args, **kwargs)
        
        # Apt sources.
        r.pc('Installing apt sources.')
        apt_sources = repos.get(APT_SOURCE, [])
        for line, fn in apt_sources:
            r.env.apt_source = line
            r.env.apt_fn = fn
            r.sudo("sh -c 'echo \"{apt_source}\" > {apt_fn}'")
        
        # Apt keys.
        r.pc('Installing apt keys.')
        apt_keys = repos.get(APT_KEY, [])
        for parts in apt_keys:
            if isinstance(parts, tuple):
                assert len(parts) == 2
                key_server, key_value = parts
                r.env.apt_key_server = key_server
                r.env.apt_key_value = key_value
                r.sudo("apt-key adv --keyserver {apt_key_server} --recv-key {apt_key_value}")
            else:
                assert isinstance(parts, basestring)
                r.env.apt_key_url = parts
                r.sudo('wget {apt_key_url} -O - | apt-key add -')
        
        r.pc('Installing repositories.')
        for repo_type in [APT]:
            if repo_type not in repos:
                continue
            repo_lst = repos[repo_type]
            if repo_type is APT:
                for repo_name in repo_lst:
                    r.env.repo_name = repo_name
                    r.sudo('add-apt-repository -y {repo_name}')
                r.sudo('DEBIAN_FRONTEND=noninteractive apt-get update -yq')
            else:
                raise NotImplementedError, 'Unsupported repository type: %s' % repo_type

    @task
    def list_required(self, type=None, service=None): # pylint: disable=redefined-builtin
        """
        Displays all packages required by the current role
        based on the documented services provided.
        """
        from burlap.common import (
            required_system_packages,
            required_python_packages,
            required_ruby_packages,
        )
        service = (service or '').strip().upper()
        type = (type or '').lower().strip()
        assert not type or type in PACKAGE_TYPES, \
            'Unknown package type: %s' % (type,)
        packages_set = set()
        packages = []
        version = self.os_version
        
        for _service, satchel in self.all_other_enabled_satchels.items():
                
            _service = _service.strip().upper()
            if service and service != _service:
                continue
                
            _new = []
            
            if not type or type == SYSTEM:
                
                #TODO:deprecated, remove
                _new.extend(required_system_packages.get(
                    _service, {}).get((version.distro, version.release), []))
                    
                try:
                    _pkgs = satchel.packager_system_packages
                    if self.verbose:
                        print('pkgs:')
                        pprint(_pkgs, indent=4)
                    for _key in [(version.distro, version.release), version.distro]:
                        if self.verbose:
                            print('checking key:', _key)
                        if _key in _pkgs:
                            if self.verbose:
                                print('satchel %s requires:' % satchel, _pkgs[_key])
                            _new.extend(_pkgs[_key])
                            break
                except AttributeError:
                    pass
                    
            if not type or type == PYTHON:
                
                #TODO:deprecated, remove
                _new.extend(required_python_packages.get(
                    _service, {}).get((version.distro, version.release), []))
                
                try:
                    _pkgs = satchel.packager_python_packages
                    for _key in [(version.distro, version.release), version.distro]:
                        if _key in _pkgs:
                            _new.extend(_pkgs[_key])
                except AttributeError:
                    pass
                print('_new:', _new)
                    
            if not type or type == RUBY:
                
                #TODO:deprecated, remove
                _new.extend(required_ruby_packages.get(
                    _service, {}).get((version.distro, version.release), []))
            
            
            
    #         if not _new and verbose:
    #             print(\
    #                 'Warning: no packages found for service "%s"' % (_service,)
            for _ in _new:
                if _ in packages_set:
                    continue
                packages_set.add(_)
                packages.append(_)
        if self.verbose:
            for package in sorted(packages):
                print('package:', package)
        return packages
    
    def get_locale(self):
        version = self.os_version
        all_locale_dicts = {}
        for satchel_name, satchel in self.all_other_enabled_satchels.items():
#             print('satchel_name:',satchel_name)
            try:
                locale_dict = satchel.packager_locale.get(version.distro, {})
#                 print('locale_dict:',locale_dict)
                for _k, _v in locale_dict.items():
                    assert all_locale_dicts.get(_k, _v) == _v
                    all_locale_dicts[_k] = _v
            except AttributeError:
                pass
        return all_locale_dicts
    
    @task
    def update_locale(self):
        locale_dict = self.get_locale()
        r = self.local_renderer
        packager = self.packager
        if packager == APT:
            r.env.locale_string = ' '.join('%s=%s' % (_k, _v) for _k, _v in locale_dict.items())
            r.sudo('update-locale {locale_string}')
        elif packager == YUM:
            raise NotImplementedError
        else:
            raise Exception('Unknown packager: %s' % (packager,))
    
    @task
    def install_required_system(self):
        self.install_required(type=SYSTEM)

    @task
    def install_required(self, type=None, service=None, list_only=0, **kwargs): # pylint: disable=redefined-builtin
        """
        Installs system packages listed as required by services this host uses.
        """
        r = self.local_renderer
#         r.pc('Installing required packages.')
        list_only = int(list_only)
        type = (type or '').lower().strip()
        assert not type or type in PACKAGE_TYPES, 'Unknown package type: %s' % (type,)
        lst = []
        if type:
            types = [type]
        else:
            types = PACKAGE_TYPES
        for type in types:
            if type == SYSTEM:
                content = '\n'.join(self.list_required(type=type, service=service))
                if list_only:
                    lst.extend(_ for _ in content.split('\n') if _.strip())
                    if self.verbose:
                        print('content:', content)
                    break
                fd, fn = tempfile.mkstemp()
                fout = open(fn, 'w')
                fout.write(content)
                fout.close()
                self.install_custom(fn=fn)
            else:
                raise NotImplementedError
        return lst

    @task(precursors=['user', 'ubuntumultiverse'])
    def configure(self, **kwargs):
        
        initial_upgrade = int(kwargs.pop('initial_upgrade', 1))
        
        service = kwargs.pop('service', '')
        
        lm = self.last_manifest
        
        if isinstance(lm, list):
            lm = {'required_packages': lm}
        
        enabled_services = map(str.upper, self.genv.services)
        #for satchel_name, satchel in self.all_satchels.iteritems():
        for satchel_name, satchel in self.all_other_enabled_satchels.items():
            if hasattr(satchel, 'packager_pre_configure'):
                satchel.packager_pre_configure()
                
        self.refresh()
        if initial_upgrade and lm.initial_upgrade is None and self.env.initial_upgrade:
            self.upgrade()
            if self.env.do_reboots:
                self.reboot(wait=300, timeout=60)
            
        self.install_repositories(service=service, **kwargs)
        self.install_required(type=SYSTEM, service=service, **kwargs)
        self.install_custom(**kwargs)

class UbuntuMultiverseSatchel(Satchel):
     
    name = 'ubuntumultiverse'

    @task
    def configure(self):
        r = self.local_renderer
        if self.env.enabled:
            # Enable the multiverse so we can install select non-free packages.
            r.sudo('which sed || DEBIAN_FRONTEND=noninteractive apt-get -yq install sed')
            r.sudo('sed -i "/^# deb.*multiverse/ s/^# //" /etc/apt/sources.list')
            r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq update')
        else:
            # Disable the multiverse.
            r.sudo('sed -i "/^# // s/^# deb.*multiverse/" /etc/apt/sources.list')
            r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq update')

packager = PackagerSatchel()
umv = UbuntuMultiverseSatchel()



"""
Tests for the common module.
"""
from __future__ import print_function

import os
import tempfile
import unittest

# try:
#     import pytest
# except ImportError:
#     pass

from burlap import load_yaml_settings
from burlap.common import CMD_VAR_REGEX, CMD_ESCAPED_VAR_REGEX
from burlap.common import shellquote
from burlap.common import Satchel, env
#from burlap.common import LocalRenderer
    
class _TestSatchel(Satchel):
    
    name = 'test'
    
    def configure(self):
        pass

class CommonTests(unittest.TestCase):
    
    def setUp(self):
        self.test = _TestSatchel()
        self.test.genv.hosts = ['localhost']
        self.test.genv.host_string = self.test.genv.hosts[0]
    
    def test_shellquote(self):
        
        s = """# /etc/cron.d/anacron: crontab entries for the anacron package
    
    SHELL=/bin/bash
    PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin
    
    # minute hour day month weekday (0-6, 0 = Sunday) user command
    
    */5 * * * *   root    {command}
    """
    
        s = shellquote(s)
        
    def test_format_regex(self):
        
        assert CMD_VAR_REGEX.findall('{cmd} {host}') == ['cmd', 'host']
        
        assert CMD_VAR_REGEX.findall("{cmd} {host} | {awk_cmd} '{{ print $1 }}'") == ['cmd', 'host', 'awk_cmd']
    
        assert CMD_ESCAPED_VAR_REGEX.findall("{cmd}} {{ print hello }}") == ['{{ print hello }}']
    
        r = self.test.local_renderer
        r.env.host = 'myhost'
        r.local("getent {host} | awk '{{ print $1 }}'", dryrun=1)
        
        s = "rsync --recursive --verbose --perms --times --links --compress --copy-links {exclude_str}  ' \
            '--delete --delete-before --force {rsync_auth} {rsync_source_dir} {rsync_target_host}{rsync_target_dir}"
        assert CMD_VAR_REGEX.findall(s) == ['exclude_str', 'rsync_auth', 'rsync_source_dir', 'rsync_target_host', 'rsync_target_dir']
    
    def test_settings_include(self):
        try:
            os.makedirs('/tmp/burlap_test/roles/all')
        except OSError:
            pass
            
        try:
            os.makedirs('/tmp/burlap_test/roles/prod')
        except OSError:
            pass
            
        open('/tmp/burlap_test/roles/all/settings.yaml', 'w').write("""
only_all_param: "just in all"
overridden_by_prod: 123
overridden_by_local: slkdjflsk
    """)
        open('/tmp/burlap_test/roles/prod/settings.yaml', 'w').write("""inherits: all
overridden_by_prod: 'prod'
only_prod_param: 7891
overriden_by_include: 7892
overridden_by_local: oiuweoiruwo
#includes: [settings_include1.yaml]
includes: [settings_include2.yaml]
    """)
        open('/tmp/burlap_test/roles/prod/settings_include2.yaml', 'w').write("""
overriden_by_include: xyz
overridden_by_local: ovmxlkfsweirwio
    """)
        open('/tmp/burlap_test/roles/prod/settings_local.yaml', 'w').write("""
overridden_by_local: 'hello world'
includes: [settings_include3.yaml]
    """)
        open('/tmp/burlap_test/roles/prod/settings_include3.yaml', 'w').write("""
set_by_include3: 'some special setting'
    """)
        os.chdir('/tmp/burlap_test')
        config = load_yaml_settings(name='prod')
        
        assert config['includes'] == ['settings_include2.yaml', 'settings_include3.yaml']
        assert config['only_all_param'] == 'just in all'
        assert config['overridden_by_prod'] == 'prod'
        assert config['only_prod_param'] == 7891
        assert config['overriden_by_include'] == 'xyz'
        assert config['overridden_by_local'] == 'hello world'
        assert config['set_by_include3'] == 'some special setting'
    
    def test_renderer(self):
        
        _env = env.copy()
        try:
            
            # Confirm renderer is cached.
            r1 = self.test.local_renderer
            r2 = self.test.local_renderer
            assert r1 is r2
            
            # Confirm clear method.
            self.test.clear_local_renderer()
            r3 = self.test.local_renderer
            assert r1 is not r3
            
            r = r3
            env.clear()
            assert r.genv is env
            
            # Confirm local env var gets renderered.
            r.env.var1 = 'a'
            assert r.format('{var1}') == 'a'
            
            # Confirm global env var in local namespace gets rendered.
            env.test_var2 = 'b'
            assert r.format('{var2}') == 'b'
            
            # Confirm global env var in global namespace gets rendered.
            env.test_var2 = 'b2'
            assert r.format('{test_var2}') == 'b2'
            
            # Confirm global env var overridden in local namespace get rendered.
            env.apache_var3 = '0'
            r.env.apache_var3 = 'c'
            assert r.format('{apache_var3}') == 'c'
            
            # Confirm recursive template variables get rendered.
            r.env.some_template = '{target_value}'
            r.env.target_value = 'd'
            assert r.format('{some_template}') == 'd'
            
            class ApacheSatchel(Satchel):
                
                name = 'apache'
                
                def configure(self):
                    pass
                
            apache = ApacheSatchel()
            r = apache.local_renderer
            r.env.application_name = 'someappname'
            r.env.site = 'sitename'
            r.env.wsgi_path = '/usr/local/{apache_application_name}/src/wsgi/{apache_site}.wsgi'
            assert r.format(r.env.wsgi_path) == '/usr/local/someappname/src/wsgi/sitename.wsgi'
        finally:
            env.update(_env)
    
    def test_iter_sites(self):
        
        env.sites = {
            'site1': {'apache_ssl': False},
            'site2': {'apache_ssl': True},
        }
        
        lst = list(self.test.iter_sites())
        print('lst:', lst)
        assert len(lst) == 2
        
        lst = list(self.test.iter_sites(site='site2'))
        print('lst:', lst)
        assert len(lst) == 1
    
    def test_append(self):
        
        self.test.genv.host_string = 'localhost'
        
        _, fn = tempfile.mkstemp()
        
        text = '[{rabbit, [{loopback_users, []}]}].'
        
        self.test.append(filename=fn, text=text)
        content = open(fn).read()
        print('content0:', content)
        assert content.count(text) == 1
        
        # Confirm duplicate lines are appended.
        self.test.append(filename=fn, text=text)
        content = open(fn).read()
        print('content1:', content)
        assert content.count(text) == 1



from burlap.constants import *
from burlap.mysql import MySQLSatchel, MYSQLD_SAFE

MYSQL_ROOT_PASSWORD = 's3cr3t'

def test_set_root_password_mysqld_safe():
    mysql = MySQLSatchel()
    try:
        mysql.verbose = True
        
        # https://docs.travis-ci.com/user/ci-environment/
        # Travis-CI uses Ubuntu 14.04, which has some odd default MySQL packages that we need to remove
        # before we can install the correct ones.
        # http://askubuntu.com/a/489817/13217
#         os_ver = mysql.os_version
#         if os_ver.distro == UBUNTU and os_ver.release == '14.04':
#             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq purge mysql*')
#             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq autoremove')
#             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq autoclean')
#             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq update')
#             mysql.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq install mysql-client-core-5.5')
            
        mysql.install_packages()
        mysql.env.root_username = 'root'
        mysql.env.root_password = mysql.env.db_root_password = MYSQL_ROOT_PASSWORD
        mysql.set_root_password(method=MYSQLD_SAFE)
        ret = mysql.execute('SHOW DATABASES;', use_sudo=True)
        print('ret:', ret)
    finally:
        mysql.purge_packages()



"""
Django-specific helper utilities.
"""
from __future__ import print_function

import os
import re
import sys
import traceback
import glob
from importlib import import_module
from collections import defaultdict
from pprint import pprint

from six import StringIO

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task
from burlap.common import get_last_modified_timestamp

class DjangoSatchel(Satchel):

    # We don't use "django" as the name so as to not conflict with the official django package.    
    name = 'dj'
    
    def set_defaults(self):
        
        # This is the name of the executable to call to access Django's management features.
        self.env.manage_cmd = 'manage.py'
        
        # This is the name of your Django application.
        self.env.app_name = None
        
        # This is the import path to your Django settings file.
        self.env.settings_module = '{app_name}.settings'
        
        # The folder containing manage.py.
        self.env.project_dir = None
        
        # The folder containing manage.py on the local filesystem.
        self.env.local_project_dir = None
        
        self.env.shell_template = 'cd {project_dir}; /bin/bash -i -c \"{manage_cmd} shell;\"'
        
        # These apps will be migrated on a specific database, while faked
        # on all others.
        # This is necessary since South does not have proper support for
        # multi-database applications.
        #./manage migrate <app> --fake
        #./manage migrate --database=<database> <app>
        self.env.migrate_fakeouts = [] # [{database:<database>, app:<app>}]
        
        self.env.install_sql_path_template = '{src_dir}/{app_name}/*/sql/*'
        
        # The target version of Django to assume.
        self.env.version = (1, 6, 0)
        
        self.env.media_dirs = ['static']
        
        # The path relative to fab where the code resides.
        self.env.src_dir = 'src'
        
        self.env.manage_dir = 'src'
        
        self.env.ignore_errors = 0
        
        # Modules whose name start with one of these values will be deleted before settings are imported.
        self.env.delete_module_with_prefixes = []

    def has_database(self, name, site=None, role=None):
        settings = self.get_settings(site=site, role=role)
        return name in settings.DATABASES

    @task
    def get_settings(self, site=None, role=None):
        """
        Retrieves the Django settings dictionary.
        """
        r = self.local_renderer
        _stdout = sys.stdout
        _stderr = sys.stderr
        if not self.verbose:
            sys.stdout = StringIO()
            sys.stderr = StringIO()
        try:
            sys.path.insert(0, r.env.src_dir)
            
            # Temporarily override SITE.
            tmp_site = self.genv.SITE
            if site and site.endswith('_secure'):
                site = site[:-7]
            site = site or self.genv.SITE or self.genv.default_site
            self.set_site(site)
            
            # Temporarily override ROLE.
            tmp_role = self.genv.ROLE
            if role:
                self.set_role(role)

            try:
                # We need to explicitly delete sub-modules from sys.modules. Otherwise, reload() skips
                # them and they'll continue to contain obsolete settings.
                if r.env.delete_module_with_prefixes:
                    for name in sorted(sys.modules):
                        for prefix in r.env.delete_module_with_prefixes:
                            if name.startswith(prefix):
                                if self.verbose:
                                    print('Deleting module %s prior to re-import.' % name)
                                del sys.modules[name]
                                break
                        
                if r.env.settings_module in sys.modules:
                    del sys.modules[r.env.settings_module]
                module = import_module(r.format(r.env.settings_module))
        
                # Works as long as settings.py doesn't also reload anything.
                import imp
                imp.reload(module)
                
            except ImportError as e:
                print('Warning: Could not import settings for site "%s": %s' % (site, e), file=_stdout)
                traceback.print_exc(file=_stdout)
                #raise # breaks *_secure pseudo sites
                return
            finally:
                if tmp_site:
                    self.set_site(tmp_site)
                if tmp_role:
                    self.set_role(tmp_role)
        finally:
            sys.stdout = _stdout
            sys.stderr = _stderr
            sys.path.remove(r.env.src_dir)
        return module

    def set_db(self, name=None, site=None, role=None):
        r = self.local_renderer
        name = name or 'default'
        site = site or r.env.get('SITE') or r.genv.SITE or r.genv.default_site
        role = role or r.env.get('ROLE') or r.genv.ROLE
        settings = self.get_settings(site=site, role=role)
        assert settings, 'Unable to load Django settings for site %s.' % (site,)
        r.env.django_settings = settings
        default_db = settings.DATABASES[name]
        r.env.db_name = default_db['NAME']
        r.env.db_user = default_db['USER']
        r.env.db_host = default_db['HOST']
        r.env.db_password = default_db['PASSWORD']
        r.env.db_engine = default_db['ENGINE']
        
        if 'mysql' in r.env.db_engine.lower():
            r.env.db_type = 'mysql'
        elif 'postgres' in r.env.db_engine.lower() or 'postgis' in r.env.db_engine.lower():
            r.env.db_type = 'postgresql'
        elif 'sqlite' in r.env.db_engine.lower():
            r.env.db_type = 'sqlite'
        else:
            r.env.db_type = r.env.db_engine
        
        for k, v in r.genv.iteritems():
            if not k.startswith(self.name.lower()+'_db_'):
                continue
            print('db.kv:', k, v)
        
        return default_db

    @task
    def install_sql(self, site=None, database='default', apps=None, stop_on_error=0):
        """
        Installs all custom SQL.
        """
        #from burlap.db import load_db_set
        
        stop_on_error = int(stop_on_error)
        
        name = database
        self.set_db(name=name, site=site)

        r = self.local_renderer
        paths = glob.glob(r.format(r.env.install_sql_path_template))
        
        apps = [_ for _ in (apps or '').split(',') if _.strip()]
        if self.verbose:
            print('install_sql.apps:', apps)
        
        def cmp_paths(d0, d1):
            if d0[1] and d0[1] in d1[2]:
                return -1
            if d1[1] and d1[1] in d0[2]:
                return +1
            return cmp(d0[0], d1[0])
        
        def get_paths(t):
            """
            Returns SQL file paths in an execution order that respect dependencies.
            """
            data = [] # [(path, view_name, content)]
            for path in paths:
                #print path
                parts = path.split('.')
                if len(parts) == 3 and parts[1] != t:
                    continue
                if not path.lower().endswith('.sql'):
                    continue
                content = open(path, 'r').read()
                matches = re.findall(r'[\s\t]+VIEW[\s\t]+([a-zA-Z0-9_]+)', content, flags=re.IGNORECASE)
                #assert matches, 'Unable to find view name: %s' % (p,)
                view_name = ''
                if matches:
                    view_name = matches[0]
                data.append((path, view_name, content))
            for d in sorted(data, cmp=cmp_paths):
                yield d[0]
        
        def run_paths(paths, cmd_template, max_retries=3):
            r = self.local_renderer
            paths = list(sorted(paths))
            error_counts = defaultdict(int) # {path:count}
            terminal = set()
            if self.verbose:
                print('Checking %i paths.' % len(paths))
            while paths:
                path = paths.pop(0)
                if self.verbose:
                    print('path:', path)
                app_name = re.findall(r'/([^/]+)/sql/', path)[0]
                if apps and app_name not in apps:
                    self.vprint('skipping because app_name %s not in apps' % app_name)
                    continue
                with self.settings(warn_only=True):
                    if r.genv.is_local:
                        r.env.sql_path = path
                    else:
                        r.env.sql_path = '/tmp/%s' % os.path.split(path)[-1]
                        r.put(local_path=path, remote_path=r.env.sql_path)
                    ret = r.run_or_local(cmd_template)
                    if ret and ret.return_code:
                        
                        if stop_on_error:
                            raise Exception('Unable to execute file %s' % path)
                            
                        error_counts[path] += 1
                        if error_counts[path] < max_retries:
                            paths.append(path)
                        else:
                            terminal.add(path)
            if terminal:
                print('%i files could not be loaded.' % len(terminal), file=sys.stderr)
                for path in sorted(list(terminal)):
                    print(path, file=sys.stderr)
                print(file=sys.stderr)
        
        if self.verbose:
            print('install_sql.db_engine:', r.env.db_engine)
        
        if 'postgres' in r.env.db_engine or 'postgis' in r.env.db_engine:
            run_paths(
                paths=get_paths('postgresql'),
                cmd_template="psql --host={db_host} --user={db_user} -d {db_name} -f {sql_path}")
                        
        elif 'mysql' in r.env.db_engine:
            run_paths(
                paths=get_paths('mysql'),
                cmd_template="mysql -v -h {db_host} -u {db_user} -p'{db_password}' {db_name} < {sql_path}")
                
        else:
            raise NotImplementedError

    @task
    def createsuperuser(self, username='admin', email=None, password=None, site=None):
        """
        Runs the Django createsuperuser management command.
        """
        r = self.local_renderer
        self.set_site_specifics(site)
        r.env.db_createsuperuser_username = username
        r.env.db_createsuperuser_email = email or username
        r.run('export SITE={SITE}; export ROLE={ROLE}; '
            'cd {project_dir}; {manage_cmd} createsuperuser '
            '--username={db_createsuperuser_username} --email={db_createsuperuser_email}')

    @task
    def loaddata(self, path, site=None):
        """
        Runs the Dango loaddata management command.
        
        By default, runs on only the current site.
        
        Pass site=all to run on all sites.
        """
        site = site or self.env.SITE
        r = self.local_renderer
        r.env._loaddata_path = path
        for site, site_data in self.iter_sites(site=site, no_secure=True):
            try:
                self.set_db(site=site)
                r.env.SITE = site
                r.sudo('export SITE={SITE}; export ROLE={ROLE}; '
                    'cd {project_dir}; '
                    '{manage_cmd} loaddata {_loaddata_path}')
            except KeyError:
                pass

    @task
    def manage(self, cmd, *args, **kwargs):
        """
        A generic wrapper around Django's manage command.
        """
        r = self.local_renderer
        environs = kwargs.pop('environs', '').strip()
        if environs:
            environs = ' '.join('export %s=%s;' % tuple(_.split('=')) for _ in environs.split(','))
            environs = ' ' + environs + ' '
        r.env.cmd = cmd
        r.env.SITE = r.genv.SITE or r.genv.default_site
        r.env.args = ' '.join(map(str, args))
        r.env.kwargs = ' '.join(
            ('--%s' % _k if _v in (True, 'True') else '--%s=%s' % (_k, _v))
            for _k, _v in kwargs.iteritems())
        r.env.environs = environs
        r.run_or_local('export SITE={SITE}; export ROLE={ROLE};{environs} cd {project_dir}; {manage_cmd} {cmd} {args} {kwargs}')

    @task
    def manage_all(self, *args, **kwargs):
        """
        Runs manage() across all unique site default databases.
        """
        for site, site_data in self.iter_unique_databases(site='all'):
            if self.verbose:
                print('-'*80, file=sys.stderr)
                print('site:', site, file=sys.stderr)
            if self.env.available_sites_by_host:
                hostname = self.current_hostname
                sites_on_host = self.env.available_sites_by_host.get(hostname, [])
                if sites_on_host and site not in sites_on_host:
                    self.vprint('skipping site:', site, sites_on_host, file=sys.stderr)
                    continue
            self.manage(*args, **kwargs)

    def load_django_settings(self):
        """
        Loads Django settings for the current site and sets them so Django internals can be run.
        """
    
        #TODO:remove this once bug in django-celery has been fixed
        os.environ['ALLOW_CELERY'] = '0'
    
        #os.environ.setdefault("DJANGO_SETTINGS_MODULE", "dryden_site.settings")
    
        # In Django >= 1.7, fixes the error AppRegistryNotReady: Apps aren't loaded yet
        try:
            from django.core.wsgi import get_wsgi_application
            application = get_wsgi_application()
        except (ImportError, RuntimeError):
            traceback.print_exc()
    
        # Load Django settings.
        settings = self.get_settings()
        try:
            from django.contrib import staticfiles
            from django.conf import settings as _settings
            for k, v in settings.__dict__.iteritems():
                setattr(_settings, k, v)
        except (ImportError, RuntimeError):
            traceback.print_exc()
            
        return settings
    
    def iter_static_paths(self, ignore_import_error=False):
        self.load_django_settings()
        from django.contrib.staticfiles import finders, storage
        for finder in finders.get_finders():
            for _n, _s in finder.storages.iteritems():
                yield _s.location
    
    def iter_app_directories(self, ignore_import_error=False):
        settings = self.load_django_settings()
        if not settings:
            return
        
        for app in settings.INSTALLED_APPS:
            try:
                mod = import_module(app)
            except ImportError:
                if ignore_import_error:
                    continue
                else:
                    raise
            yield app, os.path.dirname(mod.__file__)
    
    def iter_south_directories(self, *args, **kwargs):
        for app_name, base_app_dir in self.iter_app_directories(*args, **kwargs):
            migrations_dir = os.path.join(base_app_dir, 'migrations')
            if not os.path.isdir(migrations_dir):
                continue
            yield app_name, migrations_dir
    
    def iter_migrations(self, d, *args, **kwargs):
        for fn in sorted(os.listdir(d)):
            if fn.startswith('_') or not fn.endswith('.py'):
                continue
            fqfn = os.path.join(d, fn)
            if not os.path.isfile(fqfn):
                continue
            yield fn
    
    def iter_unique_databases(self, site=None):
        r = self.local_renderer
        prior_database_names = set()
        for site, site_data in self.iter_sites(site=site, no_secure=True):
            self.set_db(site=site)
            key = (r.env.db_name, r.env.db_user, r.env.db_host, r.env.db_engine)
            if key in prior_database_names:
                continue
            prior_database_names.add(key)
            r.env.SITE = site
            yield site, site_data
    
    @task
    def shell(self):
        """
        Opens a Django focussed Python shell.
        Essentially the equivalent of running `manage.py shell`.
        """
        r = self.local_renderer
        if '@' in self.genv.host_string:
            r.env.shell_host_string = self.genv.host_string
        else:
            r.env.shell_host_string = '{user}@{host_string}'
        r.env.shell_default_dir = self.genv.shell_default_dir_template
        r.env.shell_interactive_djshell_str = self.genv.interactive_shell_template
        r.run_or_local('ssh -t -i {key_filename} {shell_host_string} "{shell_interactive_djshell_str}"')
        
    @task
    def syncdb(self, site=None, all=0, database=None, ignore_errors=1): # pylint: disable=redefined-builtin
        """
        Runs the standard Django syncdb command for one or more sites.
        """
        r = self.local_renderer
        
        ignore_errors = int(ignore_errors)
        
        r.env.db_syncdb_all_flag = '--all' if int(all) else ''
        
        r.env.db_syncdb_database = ''
        if database:
            r.env.db_syncdb_database = ' --database=%s' % database

        for site, site_data in r.iter_unique_databases(site=site):
            r.env.SITE = site
            with self.settings(warn_only=ignore_errors):
                r.run_or_local(
                    'export SITE={SITE}; export ROLE={ROLE}; cd {project_dir}; '
                    '{manage_cmd} syncdb --noinput {db_syncdb_all_flag} {db_syncdb_database}')

    @task
    def migrate(self, app='', migration='', site=None, fake=0, ignore_errors=0, skip_databases=None, database=None, migrate_apps='', delete_ghosts=1):
        """
        Runs the standard South migrate command for one or more sites.
        """
    #     Note, to pass a comma-delimted list in a fab command, escape the comma with a back slash.
    #         
    #         e.g.
    #         
    #             fab staging dj.migrate:migrate_apps=oneapp\,twoapp\,threeapp
        
        r = self.local_renderer
        
        ignore_errors = int(ignore_errors)
        
        delete_ghosts = int(delete_ghosts)
        
        post_south = tuple(r.env.version) >= (1, 7, 0)
        
        if tuple(r.env.version) >= (1, 9, 0):
            delete_ghosts = 0
        
        skip_databases = (skip_databases or '')
        if isinstance(skip_databases, basestring):
            skip_databases = [_.strip() for _ in skip_databases.split(',') if _.strip()]
        
        migrate_apps = migrate_apps or ''    
        migrate_apps = [
            _.strip().split('.')[-1]
            for _ in migrate_apps.strip().split(',')
            if _.strip()
        ]
        if app:
            migrate_apps.append(app)
    
        #print('ignore_errors:', ignore_errors)
        
        r.env.migrate_migration = migration or ''
    #     print('r.env.migrate_migration:', r.env.migrate_migration)
        r.env.migrate_fake_str = '--fake' if int(fake) else ''
        r.env.migrate_database = '--database=%s' % database if database else ''
        r.env.migrate_merge = '--merge' if not post_south else ''
        r.env.delete_ghosts = '--delete-ghost-migrations' if delete_ghosts and not post_south else ''
        for site, site_data in self.iter_unique_databases(site=site):
    #         print('-'*80, file=sys.stderr)
    #         print('site:', site, file=sys.stderr)
            
            if self.env.available_sites_by_host:
                hostname = self.current_hostname
                sites_on_host = self.env.available_sites_by_host.get(hostname, [])
                if sites_on_host and site not in sites_on_host:
    #                 print('skipping site:', site, sites_on_host, file=sys.stderr)
                    continue
            
    #         print('migrate_apps:', migrate_apps, file=sys.stderr)
            if not migrate_apps:
                migrate_apps.append(' ')
                
            for app in migrate_apps:
    #             print('app:', app)
                r.env.migrate_app = app
    #             print('r.env.migrate_app:', r.env.migrate_app)
                r.env.SITE = site
                with self.settings(warn_only=ignore_errors):
                    r.run_or_local(
                        'export SITE={SITE}; export ROLE={ROLE}; cd {project_dir}; '
                        '{manage_cmd} migrate --noinput {migrate_merge} --traceback '
                        '{migrate_database} {delete_ghosts} {migrate_app} {migrate_migration} '
                        '{migrate_fake_str}')

    @task
    def migrate_all(self, *args, **kwargs):
        kwargs['site'] = 'all'
        return self.migrate(*args, **kwargs)

    @task
    def truncate(self, app):
        assert self.genv.SITE, 'This should only be run for a specific site.'
        r = self.local_renderer
        r.env.app = app
        r.run('rm -f {app_dir}/{app}/migrations/*.py')
        r.run('rm -f {app_dir}/{app}/migrations/*.pyc')
        r.run('touch {app_dir}/{app}/migrations/__init__.py')
        r.run('export SITE={SITE}; export ROLE={ROLE}; cd {app_dir}; {manage_cmd} schemamigration {app} --initial')
        r.run('export SITE={SITE}; export ROLE={ROLE}; cd {app_dir}; {manage_cmd} migrate {app} --fake')

    @task
    def manage_async(self, command='', name='process', site=ALL, exclude_sites='', end_message='', recipients=''):
        """
        Starts a Django management command in a screen.
        
        Parameters:
            
            command :- all arguments passed to `./manage` as a single string
            
            site :- the site to run the command for (default is all)
            
        Designed to be ran like:
        
            fab <role> dj.manage_async:"some_management_command --force"

        """
        exclude_sites = exclude_sites.split(':')
        r = self.local_renderer
        for site, site_data in self.iter_sites(site=site, no_secure=True):
            if site in exclude_sites:
                continue
            r.env.SITE = site
            r.env.command = command
            r.env.end_email_command = ''
            r.env.recipients = recipients or ''
            r.env.end_email_command = ''
            if end_message:
                end_message = end_message + ' for ' + site
                end_message = end_message.replace(' ', '_')
                r.env.end_message = end_message
                r.env.end_email_command = r.format('{manage_cmd} send_mail --subject={end_message} --recipients={recipients}')
            r.env.name = name.format(**r.genv)
            r.run(
                'screen -dmS {name} bash -c "export SITE={SITE}; '\
                'export ROLE={ROLE}; cd {project_dir}; '\
                '{manage_cmd} {command} --traceback; {end_email_command}"; sleep 3;')

    def record_manifest(self):
        manifest = super(DjangoSatchel, self).record_manifest()
        
        latest_timestamp = -1e9999999999999999
        for path in self.iter_static_paths():
            if self.verbose:
                print('checking timestamp of path:', path)
            latest_timestamp = max(
                latest_timestamp,
                get_last_modified_timestamp(path) or latest_timestamp)
        if self.verbose:
            print('latest_timestamp:', latest_timestamp)
        manifest['latest_timestamp'] = latest_timestamp

        data = {} # {app: latest_migration_name}
        for app_name, _dir in self.iter_app_directories():
            migration_dir = os.path.join(_dir, 'migrations')
            if not os.path.isdir(migration_dir):
                continue
            for migration_name in self.iter_migrations(migration_dir):
                data[app_name] = migration_name
        if self.verbose:
            print('%s.migrations:' % self.name)
            pprint(data, indent=4)
        manifest['migrations'] = data

        return manifest
    
    @task(precursors=['packager''pip'])
    def configure_media(self, *args, **kwargs):
        r = self.local_renderer
        assert r.env.local_project_dir
        r.local('cd {local_project_dir}; {manage_cmd} collectstatic --noinput')
        
    @task(precursors=['packager', 'apache', 'pip', 'tarball', 'postgresql', 'mysql'])
    def configure_migrations(self):
        last = self.last_manifest.migrations or {}
        current = self.current_manifest.get('migrations') or {}
        migrate_apps = []
        if last and current:
            if self.verbose:
                print('djangomigrations.last:', last)
                print('djangomigrations.current:', current)
            for app_name in current:
                if current[app_name] != last.get(app_name):
                    migrate_apps.append(app_name)
        if migrate_apps:
            # Note, Django's migrate command doesn't support multiple app name arguments
            # with all options, so we run it separately for each app.
            for app in migrate_apps:
                self.migrate(app=app, ignore_errors=self.env.ignore_errors)

    @task(precursors=['packager'])
    def configure(self, *args, **kwargs):
        self.configure_media()
        self.configure_migrations()

dj = DjangoSatchel()



# import pytest
# 
# 
# pytestmark = pytest.mark.network
# 
# 
# def test_require_nginx_server():
#     try:
#         from burlap.require.nginx import server
#         server()
#     finally:
#         uninstall_nginx()
# 
# 
# @pytest.yield_fixture
# def nginx_server():
#     from burlap.require.nginx import server
#     server()
#     yield
#     uninstall_nginx()
# 
# 
# def uninstall_nginx():
#     from burlap.system import distrib_family
#     family = distrib_family()
#     if family == 'debian':
#         from burlap.require.deb import nopackage
#         nopackage('nginx')
# 
# 
# def test_site_disabled(nginx_server):
# 
#     from burlap.require.nginx import disabled as require_nginx_site_disabled
#     from burlap.files import is_link
# 
#     require_nginx_site_disabled('default')
#     assert not is_link('/etc/nginx/sites-enabled/default')
# 
# 
# def test_site_enabled(nginx_server):
# 
#     from burlap.require.nginx import enabled as require_nginx_site_enabled
#     from burlap.files import is_link
# 
#     require_nginx_site_enabled('default')
#     assert is_link('/etc/nginx/sites-enabled/default')



from __future__ import print_function

import os
import sys

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task
    
class ApacheSatchel(ServiceSatchel):
    
    name = 'apache'
    
    post_deploy_command = 'reload'
    
    templates = [
        '{site_template}',
    ]
    
    @property
    def packager_system_packages(self):
        
        mod_lst = []
        
        if self.env.modevasive_enabled:
            mod_lst.append('libapache2-mod-evasive')
            
        if self.env.modsecurity_enabled:
            mod_lst.append('libapache2-modsecurity')
            
        if self.env.modrpaf_enabled:
            mod_lst.append('libapache2-mod-rpaf')
            
        if self.env.visitors_enabled:
            #TODO:fix? package removed in Ubuntu 16?
            mod_lst.append('visitors')
            
        return {
            FEDORA: ['httpd'] + mod_lst,
            UBUNTU: ['apache2'] + mod_lst,
            (UBUNTU, '12.04'): ['apache2', 'libapache2-mod-wsgi'] + mod_lst,
            (UBUNTU, '12.10'): ['apache2', 'libapache2-mod-wsgi'] + mod_lst,
            (UBUNTU, '14.04'): ['apache2', 'libapache2-mod-wsgi', 'apache2-utils'] + mod_lst,
            (UBUNTU, '14.10'): ['apache2', 'libapache2-mod-wsgi', 'apache2-utils'] + mod_lst,
            (UBUNTU, '16.04'): ['apache2', 'libapache2-mod-wsgi', 'apache2-utils'] + mod_lst,
            (UBUNTU, '16.10'): ['apache2', 'libapache2-mod-wsgi', 'apache2-utils'] + mod_lst,
        }
    
    def set_defaults(self):
        
        self.env.service_commands = {
#             START:{
#                 UBUNTU: 'service network-manager start',
#             },
#             STOP:{
#                 UBUNTU: 'service network-manager stop',
#             },
#             DISABLE:{
#                 UBUNTU: 'chkconfig network-manager off',
#             },
#             ENABLE:{
#                 UBUNTU: 'chkconfig network-manager on',
#             },
#             RESTART:{
#                 UBUNTU: 'service network-manager restart',
#             },
#             STATUS:{
#                 UBUNTU: 'service network-manager status',
#             },
            START:{
                FEDORA: 'systemctl start httpd.service',
                UBUNTU: 'service apache2 start',
            },
            STOP:{
                FEDORA: 'systemctl stop httpd.service',
                UBUNTU: 'service apache2 stop',
            },
            DISABLE:{
                FEDORA: 'systemctl disable httpd.service',
                UBUNTU: 'chkconfig apache2 off',
                (UBUNTU, '14.04'): 'update-rc.d -f apache2 remove',
            },
            ENABLE:{
                FEDORA: 'systemctl enable httpd.service',
                UBUNTU: 'chkconfig apache2 on',
                (UBUNTU, '14.04'): 'update-rc.d apache2 defaults',
            },
            RELOAD:{
                FEDORA: 'systemctl reload httpd.service',
                UBUNTU: 'service apache2 reload',
            },
            RESTART:{
                FEDORA: 'systemctl restart httpd.service',
                #UBUNTU: 'service apache2 restart',
                # Note, the sleep 5 is necessary because the stop/start appears to
                # happen in the background but gets aborted if Fabric exits before
                # it completes.
                UBUNTU: 'service apache2 restart; sleep 3',
            },
        }
        
        # An Apache-conf file and filename friendly string that uniquely identifies
        # your web application.
        self.env.application_name = None
        
        # The Jinja-formatted template file used to render site configurations.
        self.env.site_template = 'apache/apache_site.template.conf'
        
        self.env.error_log = '/var/log/apache2/error.log'
        self.env.log_level = 'warn'
        
        self.env.auth_basic = False
        self.env.auth_basic_authuserfile = '{apache_docroot}/.htpasswd_{apache_site}'
        self.env.auth_basic_users = [] # [(user,password)]
        
        # If true, activates a rewrite rule that causes domain.com to redirect
        # to www.domain.com.
        self.env.enforce_subdomain = True
        
        self.env.ssl = True
        self.env.ssl_chmod = 440
        
        # A list of path patterns that should have HTTPS enforced.
        self.env.ssl_secure_paths_enforce = True
        self.env.ssl_secure_paths = ['/admin/(.*)']
        
        self.env.web_user = 'www-data'
        self.env.web_group = 'www-data'
        self.env.wsgi_user = 'www-data'
        self.env.wsgi_group = 'www-data'
        self.env.chmod = 775
        
        self.env.mods_enabled = ['rewrite', 'wsgi', 'ssl']
        
        # The value of the Apache's ServerName field. Usually should be set
        # to the domain.
        self.env.server_name = None
        
        self.env.server_aliases_template = ''
        
        self.env.docroot = '/usr/local/{apache_application_name}'
        self.env.ports_path = '{apache_root}/ports.conf'
        self.env.ssl_path = '{apache_root}/ssl'
        
        self.env.domain_with_sub_template = ''
        self.env.domain_without_sub_template = ''
        self.env.domain_with_sub = None
        self.env.domain_without_sub = None
        
        self.env.wsgi_enabled = False
        self.env.wsgi_template = 'django/django.template.wsgi'
        self.env.wsgi_python_path = None
        self.env.wsgi_scriptalias = None
        self.env.wsgi_server_memory_gb = 8
        self.env.wsgi_processes = 5
        self.env.wsgi_threads = 15
        
        self.env.domain_redirect_templates = [] # [(wrong_domain,right_domain)]
        self.env.domain_redirects = [] # [(wrong_domain,right_domain)]
        
        self.env.extra_rewrite_rules = ''
        
        self.env.modrpaf_enabled = False
        
        self.env.visitors_enabled = False
        
        self.env.modevasive_enabled = False
        self.env.modevasive_DOSEmailNotify = 'admin@localhost'
        self.env.modevasive_DOSPageInterval = 1 # seconds
        self.env.modevasive_DOSPageCount = 2
        self.env.modevasive_DOSSiteCount = 50
        self.env.modevasive_DOSSiteInterval = 1 # seconds
        self.env.modevasive_DOSBlockingPeriod = 10 # seconds
        
        self.env.modsecurity_enabled = False
        self.env.modsecurity_download_url = 'https://github.com/SpiderLabs/owasp-modsecurity-crs/tarball/master'
        
        # OS specific default settings.
        self.env.specifics = type(self.genv)()
        self.env.specifics[LINUX] = type(self.genv)()
        
        self.env.specifics[LINUX][FEDORA] = type(self.genv)()
        self.env.specifics[LINUX][FEDORA].root = '/etc/httpd'
        self.env.specifics[LINUX][FEDORA].conf = '/etc/httpd/conf/httpd.conf'
        self.env.specifics[LINUX][FEDORA].sites_available = '/etc/httpd/sites-available'
        self.env.specifics[LINUX][FEDORA].sites_enabled = '/etc/httpd/sites-enabled'
        self.env.specifics[LINUX][FEDORA].log_dir = '/var/log/httpd'
        self.env.specifics[LINUX][FEDORA].pid = '/var/run/httpd/httpd.pid'
        
        self.env.specifics[LINUX][UBUNTU] = type(self.genv)()
        self.env.specifics[LINUX][UBUNTU].root = '/etc/apache2'
        self.env.specifics[LINUX][UBUNTU].conf = '/etc/apache2/httpd.conf'
        self.env.specifics[LINUX][UBUNTU].sites_available = '/etc/apache2/sites-available'
        self.env.specifics[LINUX][UBUNTU].sites_enabled = '/etc/apache2/sites-enabled'
        self.env.specifics[LINUX][UBUNTU].log_dir = '/var/log/apache2'
        self.env.specifics[LINUX][UBUNTU].pid = '/var/run/apache2/apache2.pid'
        
        self.env.ssl_certificates = None
        self.env.ssl_certificates_templates = []
        
        self.env.ignored_template_variables = [
            'APACHE_LOG_DIR',
            'GLOBAL',
            'DOCUMENT_ROOT',
            'SCRIPT_FILENAME',
            'GROUP',
        ]
        
        # The local and remote relative directory where the SSL certificates are stored.
        self.env.ssl_dir_local = 'ssl'
        
        # An optional segment to insert into the domain, customizable by role.
        # Useful for easily keying domain-local.com/domain-dev.com/domain-staging.com.
        self.env.locale = ''
        
        self.env.sync_sets = {} # {name:[dict(local_path='static/', remote_path='$AWS_BUCKET:/')]}
        
        # This will be appended to the custom Apache configuration file.
        self.env.httpd_conf_append = []
    
    @task
    def enable_mod(self, name):
        self.sudo('a2enmod %s' % name)
    
    @task
    def disable_mod(self, name):
        self.sudo('a2dismod %s' % name)
        
    @task
    def enable_site(self, name):
        self.sudo('a2ensite %s' % name)
        
    @task
    def disable_site(self, name):
        self.sudo('a2dissite %s' % name)
    
    @task
    def optimize_wsgi_processes(self):
        """
        Based on the number of sites per server and the number of resources on the server,
        calculates the optimal number of processes that should be allocated for each WSGI site.
        """
        #self.env.wsgi_processes = 5
        self.env.wsgi_server_memory_gb = 8
        
        verbose = self.verbose
        
        all_sites = list(self.iter_sites(site=ALL, setter=self.set_site_specifics))
        
        #(current_mem/current_sites)/current_process = ()
        #(16/x)/(8/16) = y
        #(16/x)*(16/8) = y
        #(16*16)/(num_sites*8) = y
        
#     @task
#     def visitors(self, force=0):
#         """
#         Generates an Apache access report using the Visitors command line tool.
#         Requires the APACHE2_VISITORS service to be enabled for the current host.
#         """
#         if not int(force):
#             assert ApacheVisitors.name.upper() in self.genv.services or ApacheVisitors.name.lower() in self.genv.services, \
#                 'Visitors has not been configured for this host.'
#         self.run('visitors -o text /var/log/apache2/%(apache_application_name)s-access.log* | less' % self.genv)
    
    def create_local_renderer(self):
        """
        Instantiates a new local renderer.
        Override this to do any additional initialization.
        """
        r = super(ApacheSatchel, self).create_local_renderer()
        
        # Dynamically set values based on target operating system.
        os_version = self.os_version
        apache_specifics = r.genv.apache_specifics[os_version.type][os_version.distro]
        r.env.update(apache_specifics)
#         r.env.root = apache_specifics.root
#         r.env.conf = apache_specifics.conf
#         r.env.sites_available = apache_specifics.sites_available
#         r.env.sites_enabled = apache_specifics.sites_enabled
#         r.env.log_dir = apache_specifics.log_dir
#         r.env.pid = apache_specifics.pid
        #r.env.ports = r.env.ports_path % self.genv
        #r.env.ssl_dir = r.env.ssl_path % self.genv
        
        return r
        
#     def iter_certificates(self):
#         if self.verbose:
#             print('apache_ssl_domain:', self.genv.apache_ssl_domain, file=sys.stderr)
#         for cert_type, cert_file_template in self.genv.apache_ssl_certificates_templates:
#             if self.verbose:
#                 print('cert_type, cert_file_template:', cert_type, cert_file_template, file=sys.stderr)
#             _local_cert_file = os.path.join(self.genv.apache_ssl_dir_local, cert_file_template % self.genv)
#             local_cert_file = self.find_template(_local_cert_file)
#             assert local_cert_file, 'Unable to find local certificate file: %s' % (_local_cert_file,)
#             remote_cert_file = os.path.join(self.genv.apache_ssl_dir, cert_file_template % self.genv)
#             yield cert_type, local_cert_file, remote_cert_file
#     
#     @task
#     def install_ssl(self, site=ALL):
#         from burlap.common import iter_sites
#         verbose = self.verbose
#         
#         for site, site_data in iter_sites(site=site, setter=self.set_site_specifics):
#             
#             site_secure = site+'_secure'
#             if site_secure not in self.genv.sites:
#                 continue
#             self.set_site_specifics(site_secure)
#         
#             self.sudo_or_dryrun('mkdir -p %(apache_ssl_dir)s' % self.genv)
#             
#             if self.genv.apache_ssl:
#                 for cert_type, local_cert_file, remote_cert_file in self.iter_certificates():
#                     if verbose:
#                         print('='*80)
#                         print('Installing certificate %s...' % (remote_cert_file,))
#                     self.put_or_dryrun(
#                         local_path=local_cert_file,
#                         remote_path=remote_cert_file,
#                         use_sudo=True)
#         
#         self.sudo_or_dryrun('mkdir -p %(apache_ssl_dir)s' % self.genv)
#         self.sudo_or_dryrun('chown -R %(apache_web_user)s:%(apache_web_group)s %(apache_ssl_dir)s' % self.genv)
#         self.sudo_or_dryrun('chmod -R %(apache_ssl_chmod)s %(apache_ssl_dir)s' % self.genv)
    
    @task
    def install_auth_basic_user_file(self, site=None):
        """
        Installs users for basic httpd auth.
        """        
        r = self.local_renderer
                
        hostname = self.current_hostname
        
        target_sites = self.genv.available_sites_by_host.get(hostname, None)
        
        for site, site_data in self.iter_sites(site=site, setter=self.set_site_specifics):
            if self.verbose:
                print('~'*80, file=sys.stderr)
                print('Site:', site, file=sys.stderr)
                print('env.apache_auth_basic:', r.env.auth_basic, file=sys.stderr)
            
            # Only load site configurations that are allowed for this host.
            if target_sites is not None:
                assert isinstance(target_sites, (tuple, list))
                if site not in target_sites:
                    continue
            
            if not r.env.auth_basic:
                continue
            
            assert r.env.auth_basic_users, 'No apache auth users specified.'
            for username, password in r.env.auth_basic_users:
                r.env.auth_basic_username = username
                r.env.auth_basic_password = password
                if self.files.exists(r.env.auth_basic_authuserfile):
                    r.sudo('htpasswd -b {auth_basic_authuserfile} {auth_basic_username} {auth_basic_password}')
                else:
                    r.sudo('htpasswd -b -c {auth_basic_authuserfile} {auth_basic_username} {auth_basic_password}')
    
    @task
    def install_auth_basic_user_file_all(self):
        self.install_auth_basic_user_file(site='all')
    
    @task
    def view_error_log(self):
        self.run('tail -f {apache_error_log}')

    @task
    def sync_media(self, sync_set=None, clean=0, iter_local_paths=0):
        """
        Uploads select media to an Apache accessible directory.
        """
        
        #from burlap.dj import render_remote_paths
        
        # Ensure a site is selected.
        self.genv.SITE = self.genv.SITE or self.genv.default_site
        
#         apache.get_apache_settings()
        
        #render_remote_paths()
        
        r = self.local_renderer
        
        clean = int(clean)
        print('Getting site data for %s...' % self.genv.SITE)
        
        self.set_site_specifics(self.genv.SITE)
        #site_data = self.genv.sites[self.genv.SITE]
        #self.genv.update(site_data)
        
        sync_sets = r.env.sync_sets
        if sync_set:
            sync_sets = [sync_set]
        
        ret_paths = []
        for sync_set in sync_sets:
            for paths in r.env.sync_sets[sync_set]:
                #print 'paths:',paths
                r.env.sync_local_path = os.path.abspath(paths['local_path'] % self.genv)
                if paths['local_path'].endswith('/') and not r.env.sync_local_path.endswith('/'):
                    r.env.sync_local_path += '/'
                    
                if iter_local_paths:
                    ret_paths.append(r.env.sync_local_path)
                    continue
                    
                r.env.sync_remote_path = paths['remote_path'] % self.genv
                
                if clean:
                    r.sudo('rm -Rf {apache_sync_remote_path}') 
                
                print('Syncing %s to %s...' % (r.env.sync_local_path, r.env.sync_remote_path))
                
                r.env.tmp_chmod = paths.get('chmod', r.env.chmod)
                #with settings(warn_only=True):
                r.sudo('mkdir -p {apache_sync_remote_path}')
                r.sudo('chmod -R {apache_tmp_chmod} {apache_sync_remote_path}')
                r.local('rsync -rvz --progress --recursive --no-p --no-g '
                    '--rsh "ssh -o StrictHostKeyChecking=no -i {key_filename}" {apache_sync_local_path} {user}@{host_string}:{apache_sync_remote_path}')
                r.sudo('chown -R {apache_web_user}:{apache_web_group} {apache_sync_remote_path}')
                
        if iter_local_paths:
            return ret_paths

    def get_media_timestamp(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        from burlap.common import get_last_modified_timestamp
        data = 0
        for path in self.sync_media(iter_local_paths=1):
            data = min(data, get_last_modified_timestamp(path) or data)
        #TODO:hash media names and content
        if self.verbose:
            print('date:', data)
        return data

    @task
    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        manifest = super(ApacheSatchel, self).record_manifest()
        manifest['available_sites'] = self.genv.available_sites
        manifest['available_sites_by_host'] = self.genv.available_sites_by_host
        manifest['media_timestamp'] = self.get_media_timestamp()
        return manifest

    @task
    def configure_modevasive(self):
        """
        Installs the mod-evasive Apache module for combating DDOS attacks.
        
        https://www.linode.com/docs/websites/apache-tips-and-tricks/modevasive-on-apache
        """
        r = self.local_renderer
        if r.env.modevasive_enabled:
            self.install_packages()
                
            # Write conf for each Ubuntu version since they don't conflict.
            fn = r.render_to_file('apache/apache_modevasive.template.conf')
            
            # Ubuntu 12.04
            r.put(
                local_path=fn,
                remote_path='/etc/apache2/mods-available/mod-evasive.conf',
                use_sudo=True)
            
            # Ubuntu 14.04
            r.put(
                local_path=fn,
                remote_path='/etc/apache2/mods-available/evasive.conf',
                use_sudo=True)
            
            self.enable_mod('evasive')
        else:
#             print('self.last_manifest:', self.last_manifest)
#             print('a:', self.last_manifest.apache_modevasive_enabled)
#             print('b:', self.last_manifest.modevasive_enabled)
            if self.last_manifest.modevasive_enabled:
                self.disable_mod('evasive')

    @task
    def configure_modsecurity(self):
        """
        Installs the mod-security Apache module.
        
        https://www.modsecurity.org
        """
        r = self.local_renderer
        if r.env.modsecurity_enabled and not self.last_manifest.modsecurity_enabled:
            
            self.install_packages()
            
            # Write modsecurity.conf.
            fn = self.render_to_file('apache/apache_modsecurity.template.conf')
            r.put(local_path=fn, remote_path='/etc/modsecurity/modsecurity.conf', use_sudo=True)
            
            # Write OWASP rules.
            r.env.modsecurity_download_filename = '/tmp/owasp-modsecurity-crs.tar.gz'
            r.sudo('cd /tmp; wget --output-document={apache_modsecurity_download_filename} {apache_modsecurity_download_url}')
            r.env.modsecurity_download_top = r.sudo(
                "cd /tmp; "
                "tar tzf %(apache_modsecurity_download_filename)s | sed -e 's@/.*@@' | uniq" % self.genv)
            r.sudo('cd /tmp; tar -zxvf %(apache_modsecurity_download_filename)s' % self.genv)
            r.sudo('cd /tmp; cp -R %(apache_modsecurity_download_top)s/* /etc/modsecurity/' % self.genv)
            r.sudo('mv /etc/modsecurity/modsecurity_crs_10_setup.conf.example  /etc/modsecurity/modsecurity_crs_10_setup.conf')
            
            r.sudo('rm -f /etc/modsecurity/activated_rules/*')
            r.sudo('cd /etc/modsecurity/base_rules; '
                'for f in * ; do ln -s /etc/modsecurity/base_rules/$f /etc/modsecurity/activated_rules/$f ; done')
            r.sudo('cd /etc/modsecurity/optional_rules; '
                'for f in * ; do ln -s /etc/modsecurity/optional_rules/$f /etc/modsecurity/activated_rules/$f ; done')
            
            r.env.httpd_conf_append.append('Include "/etc/modsecurity/activated_rules/*.conf"')
            
            self.enable_mod('evasive')
            self.enable_mod('headers')
        elif not self.env.modsecurity_enabled and self.last_manifest.modsecurity_enabled:
                self.disable_mod('modsecurity')
    @task
    def configure_modrpaf(self):
        """
        Installs the mod-rpaf Apache module.
        
        https://github.com/gnif/mod_rpaf
        """
        r = self.local_renderer
        if r.env.modrpaf_enabled:
            self.install_packages()
            self.enable_mod('rpaf')
        else:
            if self.last_manifest.modrpaf_enabled:
                self.disable_mod('mod_rpaf')
                
    @task
    def configure_site(self, full=1, site=None, delete_old=0):
        """
        Configures Apache to host one or more websites.
        """
        from burlap import service
        
        r = self.local_renderer
        
        print('Configuring Apache...', file=sys.stderr)
        
        site = site or self.genv.SITE
        
        if int(delete_old):
            # Delete all existing enabled and available sites.
            r.sudo('rm -f {sites_available}/*')
            r.sudo('rm -f {sites_enabled}/*')
        
        for site, site_data in self.iter_sites(site=site, setter=self.set_site_specifics):
            r = self.local_renderer

            #r.env.site = site
            if self.verbose:
                print('-'*80, file=sys.stderr)
                print('Site:', site, file=sys.stderr)
                print('-'*80, file=sys.stderr)
            
            r.env.server_name = r.format(r.env.domain_template)
            print('r.env.server_name:', r.env.server_name)
            
            # Write WSGI template
            if r.env.wsgi_enabled:
                r.pc('Writing WSGI template for site %s...' % site)
                r.env.wsgi_scriptalias = r.format(r.env.wsgi_scriptalias)
                fn = self.render_to_file(r.env.wsgi_template)
                r.env.wsgi_dir = r.env.remote_dir = os.path.split(r.env.wsgi_scriptalias)[0]
                r.sudo('mkdir -p {remote_dir}')
                r.put(local_path=fn, remote_path=r.env.wsgi_scriptalias, use_sudo=True)
            
            # Write site configuration.
            r.pc('Writing site configuration for site %s...' % site)
            from functools import partial
            genv = r.collect_genv()
            print('*'*80)
            print('apache_wsgi_scriptalias:', genv.apache_wsgi_scriptalias)
            fn = self.render_to_file(
                self.env.site_template,
                extra=genv,
                formatter=partial(r.format, ignored_variables=self.env.ignored_template_variables))
            r.env.site_conf = site+'.conf'
            r.env.site_conf_fqfn = os.path.join(r.env.sites_available, r.env.site_conf)
            r.put(local_path=fn, remote_path=r.env.site_conf_fqfn, use_sudo=True)
            
            self.enable_site(site)
            
            self.clear_local_renderer()
        
        # Enable modules.
        for mod_name in r.env.mods_enabled:
            with self.settings(warn_only=True):
                self.enable_mod(mod_name)
            
        if int(full):
            # Write master Apache configuration file.
            fn = self.render_to_file('apache/apache_httpd.template.conf')
            r.put(local_path=fn, remote_path=r.env.conf, use_sudo=True)
            
            # Write Apache listening ports configuration.
            fn = self.render_to_file('apache/apache_ports.template.conf')
            r.put(local_path=fn, remote_path=r.env.ports_path, use_sudo=True)
            
        r.sudo('chown -R {apache_web_user}:{apache_web_group} {apache_root}')

    @task(precursors=['packager', 'user', 'hostname', 'ip'])
    def configure(self):
        self.configure_modevasive()
        self.configure_modsecurity()
        self.configure_modrpaf()
        self.configure_site(full=1, site=ALL, delete_old=1)
        self.install_auth_basic_user_file(site=ALL)
        self.sync_media()
        #self.install_ssl(site=ALL)
            
apache = ApacheSatchel()



import socket


def is_valid_ipv4_address(address):
    try:
        _ = socket.inet_pton(socket.AF_INET, address)
    except AttributeError:  # no inet_pton here, sorry
        try:
            _ = socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:    # not a valid address
        return False
    return True


def is_valid_ipv6_address(address):
    try:
        _ = socket.inet_pton(socket.AF_INET6, address)
    except socket.error:    # not a valid address
        return False
    return True


def is_valid_address(address):
    return is_valid_ipv4_address(address) or is_valid_ipv6_address()


def test_interfaces_ipv4_addresses():

    from burlap.network import interfaces
    from burlap.network import address

    for interface in interfaces():
        ipv4_address = address(interface)
        assert is_valid_ipv4_address(ipv4_address)


def test_loopback_interface_exists():
    from burlap.network import interfaces
    assert 'lo' in interfaces()


def test_loopback_interface_address():
    from burlap.network import address
    assert address('lo') == '127.0.0.1'


def test_name_servers():
    from burlap.network import nameservers
    for address in nameservers():
        assert is_valid_address(address)



from __future__ import print_function

import os
import sys
import re
import json

from fabric.api import (
    settings,
    runs_once,
)

try:
    import boto
except ImportError:
    boto = None

from burlap.constants import *
from burlap import Satchel
from burlap.decorators import task

S3SYNC_PATH_PATTERN = r'(?:->)\s+([^\n]+)'

class S3Satchel(Satchel):
    
    name = 's3'

    def set_defaults(self):
        # {name:[dict(local_path='static/', remote_path='$AWS_BUCKET:/')]}
        self.env.sync_sets = {}
        self.env.media_postfix = ''
        self.env.sync_enabled = False
        self.env.s3cmd_path = '{virtualenv_bin_dir}/s3cmd'

    @task
    @runs_once
    def sync(self, sync_set, force=0, site=None, role=None):
        """
        Uploads media to an Amazon S3 bucket using s3sync.
        
        Requires s3cmd. Install with:
        
            pip install s3cmd
            
        """
        from burlap.dj import dj
        force = int(force)
        
        r = self.local_renderer
        
        r.env.sync_force_flag = ' --force ' if force else ''
        
        _settings = dj.get_settings(site=site, role=role)
        assert _settings, 'Unable to import settings.'
        for k in _settings.__dict__.iterkeys():
            if k.startswith('AWS_'):
                r.genv[k] = _settings.__dict__[k]
        
        site_data = r.genv.sites[r.genv.SITE]
        r.env.update(site_data)
        
        r.env.virtualenv_bin_dir = os.path.split(sys.executable)[0]
        
        rets = []
        for paths in r.env.sync_sets[sync_set]:
            is_local = paths.get('is_local', True)
            local_path = paths['local_path'] % r.genv
            remote_path = paths['remote_path']
            remote_path = remote_path.replace(':/', '/')
            if not remote_path.startswith('s3://'):
                remote_path = 's3://' + remote_path
            local_path = local_path % r.genv
            
            if is_local:
                #local_or_dryrun('which s3sync')#, capture=True)
                r.env.local_path = os.path.abspath(local_path)
            else:
                #run('which s3sync')
                r.env.local_path = local_path
                
            if local_path.endswith('/') and not r.env.local_path.endswith('/'):
                r.env.local_path = r.env.local_path + '/'
                
            r.env.remote_path = remote_path % r.genv
            
            print('Syncing %s to %s...' % (r.env.local_path, r.env.remote_path))
            
            # Superior Python version.
            if force:
                r.env.sync_cmd = 'put'
            else:
                r.env.sync_cmd = 'sync'
            r.local(
                'export AWS_ACCESS_KEY_ID={aws_access_key_id}; '\
                'export AWS_SECRET_ACCESS_KEY={aws_secret_access_key}; '\
                '{s3cmd_path} {sync_cmd} --progress --acl-public --guess-mime-type --no-mime-magic '\
                '--delete-removed --cf-invalidate --recursive {sync_force_flag} '\
                '{local_path} {remote_path}')
    
    @task
    def invalidate(self, *paths):
        """
        Issues invalidation requests to a Cloudfront distribution
        for the current static media bucket, triggering it to reload the specified
        paths from the origin.
        
        Note, only 1000 paths can be issued in a request at any one time.
        """
        dj = self.get_satchel('dj')
        if not paths:
            return
        # http://boto.readthedocs.org/en/latest/cloudfront_tut.html
        _settings = dj.get_settings()
        if not _settings.AWS_STATIC_BUCKET_NAME:
            print('No static media bucket set.')
            return
        if isinstance(paths, basestring):
            paths = paths.split(',')
        all_paths = map(str.strip, paths)
    #    assert len(paths) <= 1000, \
    #        'Cloudfront invalidation request limited to 1000 paths or less.'
        i = 0
        while 1:
            paths = all_paths[i:i+1000]
            if not paths:
                break
            
            c = boto.connect_cloudfront()
            rs = c.get_all_distributions()
            target_dist = None
            for dist in rs:
                print(dist.domain_name, dir(dist), dist.__dict__)
                bucket_name = dist.origin.dns_name.replace('.s3.amazonaws.com', '')
                if bucket_name == _settings.AWS_STATIC_BUCKET_NAME:
                    target_dist = dist
                    break
            if not target_dist:
                raise Exception(('Target distribution %s could not be found '
                    'in the AWS account.') \
                        % (settings.AWS_STATIC_BUCKET_NAME,))
            print('Using distribution %s associated with origin %s.' \
                % (target_dist.id, _settings.AWS_STATIC_BUCKET_NAME))
            inval_req = c.create_invalidation_request(target_dist.id, paths)
            print('Issue invalidation request %s.' % (inval_req,))
            i += 1000

    @task
    def get_or_create_bucket(self, name):
        """
        Gets an S3 bucket of the given name, creating one if it doesn't already exist.
        
        Should be called with a role, if AWS credentials are stored in role settings. e.g.
        
            fab local s3.get_or_create_bucket:mybucket
        """
        from boto.s3 import connection
        if self.dryrun:
            print('boto.connect_s3().create_bucket(%s)' % repr(name))
        else:
            conn = connection.S3Connection(
                self.genv.aws_access_key_id,
                self.genv.aws_secret_access_key
            )
            bucket = conn.create_bucket(name)
            return bucket

    @task
    def list_bucket_sizes(self):
        
        UNIT_TO_BYTES = {
            'Bytes': 1,
            'KiB': 1000,
            'MiB': 1000000,
            'GiB': 1000000000,
        }
        
        r = self.local_renderer
        
        sizes = {} # {bucket: total size}
        total_size_bytes = 0
        
        buckets = json.loads(r.local("aws s3api list-buckets --query 'Buckets[].Name'", capture=True) or '[]')
        print('buckets:', len(buckets))
        
        for bucket in buckets:
            ret = r.local("aws s3 ls --summarize --human-readable --recursive s3://%s/" % bucket, capture=True)
            matches = re.findall(r'Total Size: (?P<number>[0-9\.]+)\s(?P<unit>[a-zA-Z]+)', ret)
            print('matches:', matches)
            value, unit = matches[0]
            value = float(value) * UNIT_TO_BYTES[unit]
            total_size_bytes += value
            sizes[bucket] = value
        print()
        print('name,bytes')
        for name in sorted(sizes):
            print('%s,%s' % (name, sizes[name]))
        print('all,%s' % total_size_bytes)
    
    @task(precursors=['packager'])
    def configure(self, *args, **kwargs):
        pass
        
s3 = S3Satchel()



"""
Git
===

This module provides low-level tools for managing `Git`_ repositories.  You
should normally not use them directly but rather use the high-level wrapper
:func:`fabtools.require.git.working_copy` instead.

.. _Git: http://git-scm.com/

"""
from __future__ import print_function

from fabric.api import run
from fabric.api import sudo
from fabric.context_managers import cd

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task
from burlap.exceptions import AbortDeployment
from burlap.utils import run_as_root


def clone(remote_url, path=None, use_sudo=False, user=None):
    """
    Clone a remote Git repository into a new directory.

    :param remote_url: URL of the remote repository to clone.
    :type remote_url: str

    :param path: Path of the working copy directory.  Must not exist yet.
    :type path: str

    :param use_sudo: If ``True`` execute ``git`` with
                     :func:`fabric.operations.sudo`, else with
                     :func:`fabric.operations.run`.
    :type use_sudo: bool

    :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                 with the given user.  If ``use_sudo is False`` this parameter
                 has no effect.
    :type user: str
    """

    cmd = 'git clone --quiet %s' % remote_url
    if path is not None:
        cmd = cmd + ' %s' % path

    if use_sudo and user is None:
        run_as_root(cmd)
    elif use_sudo:
        sudo(cmd, user=user)
    else:
        run(cmd)


def add_remote(path, name, remote_url, use_sudo=False, user=None, fetch=True):
    """
    Add a remote Git repository into a directory.

    :param path: Path of the working copy directory.  This directory must exist
                 and be a Git working copy with a default remote to fetch from.
    :type path: str

    :param use_sudo: If ``True`` execute ``git`` with
                     :func:`fabric.operations.sudo`, else with
                     :func:`fabric.operations.run`.
    :type use_sudo: bool

    :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                 with the given user.  If ``use_sudo is False`` this parameter
                 has no effect.
    :type user: str

    :param name: name for the remote repository
    :type name: str

    :param remote_url: URL of the remote repository
    :type remote_url: str

    :param fetch: If ``True`` execute ``git remote add -f``
    :type fetch: bool
    """
    if path is None:
        raise ValueError("Path to the working copy is needed to add a remote")

    if fetch:
        cmd = 'git remote add -f %s %s' % (name, remote_url)
    else:
        cmd = 'git remote add %s %s' % (name, remote_url)

    with cd(path):
        if use_sudo and user is None:
            run_as_root(cmd)
        elif use_sudo:
            sudo(cmd, user=user)
        else:
            run(cmd)


def fetch(path, use_sudo=False, user=None, remote=None):
    """
    Fetch changes from the default remote repository.

    This will fetch new changesets, but will not update the contents of
    the working tree unless yo do a merge or rebase.

    :param path: Path of the working copy directory.  This directory must exist
                 and be a Git working copy with a default remote to fetch from.
    :type path: str

    :param use_sudo: If ``True`` execute ``git`` with
                     :func:`fabric.operations.sudo`, else with
                     :func:`fabric.operations.run`.
    :type use_sudo: bool

    :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                 with the given user.  If ``use_sudo is False`` this parameter
                 has no effect.
    :type user: str

    :type remote: Fetch this remote or default remote if is None
    :type remote: str
    """

    if path is None:
        raise ValueError("Path to the working copy is needed to fetch from a "
                         "remote repository.")

    if remote is not None:
        cmd = 'git fetch %s' % remote
    else:
        cmd = 'git fetch'

    with cd(path):
        if use_sudo and user is None:
            run_as_root(cmd)
        elif use_sudo:
            sudo(cmd, user=user)
        else:
            run(cmd)


def pull(path, use_sudo=False, user=None, force=False):
    """
    Fetch changes from the default remote repository and merge them.

    :param path: Path of the working copy directory.  This directory must exist
                 and be a Git working copy with a default remote to pull from.
    :type path: str

    :param use_sudo: If ``True`` execute ``git`` with
                     :func:`fabric.operations.sudo`, else with
                     :func:`fabric.operations.run`.
    :type use_sudo: bool

    :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                 with the given user.  If ``use_sudo is False`` this parameter
                 has no effect.
    :type user: str
    :param force: If ``True``, append the ``--force`` option to the command.
    :type force: bool
    """

    if path is None:
        raise ValueError("Path to the working copy is needed to pull from a "
                         "remote repository.")

    options = []
    if force:
        options.append('--force')
    options = ' '.join(options)

    cmd = 'git pull %s' % options

    with cd(path):
        if use_sudo and user is None:
            run_as_root(cmd)
        elif use_sudo:
            sudo(cmd, user=user)
        else:
            run(cmd)


def checkout(path, branch="master", use_sudo=False, user=None, force=False):
    """
    Checkout a branch to the working directory.

    :param path: Path of the working copy directory.  This directory must exist
                 and be a Git working copy.
    :type path: str

    :param branch: Name of the branch to checkout.
    :type branch: str

    :param use_sudo: If ``True`` execute ``git`` with
                     :func:`fabric.operations.sudo`, else with
                     :func:`fabric.operations.run`.
    :type use_sudo: bool

    :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                 with the given user.  If ``use_sudo is False`` this parameter
                 has no effect.
    :type user: str
    :param force: If ``True``, append the ``--force`` option to the command.
    :type force: bool
    """

    if path is None:
        raise ValueError("Path to the working copy is needed to checkout a "
                         "branch")

    options = []
    if force:
        options.append('--force')
    options = ' '.join(options)

    cmd = 'git checkout %s %s' % (branch, options)

    with cd(path):
        if use_sudo and user is None:
            run_as_root(cmd)
        elif use_sudo:
            sudo(cmd, user=user)
        else:
            run(cmd)


class GitCheckerSatchel(Satchel):
    """
    Ensures the appropriate Git branch is being deployed.
    """
    
    name = 'gitchecker'
    
    @property
    def packager_system_packages(self):
        return {
            (UBUNTU, '12.04'): ['git'],
            (UBUNTU, '14.04'): ['git'],
        }
    
    def set_defaults(self):
        self.env.branch = 'master'
    
    @task
    def check(self):
        print('Checking GIT branch...')
        branch_name = self._local('git rev-parse --abbrev-ref HEAD', capture=True).strip()
        if not self.env.branch == branch_name:
            raise AbortDeployment(
                'Expected branch "%s" but see branch "%s".' % (self.env.branch, branch_name))
    
    def record_manifest(self):
        self.check()
        return super(GitCheckerSatchel, self).record_manifest()
    
    @task
    def configure(self):
        pass
    

class GitTrackerSatchel(Satchel):
    """
    Tracks changes between Git commits.
    """
    
    name = 'gittracker'
    
    @property
    def packager_system_packages(self):
        return {
            (UBUNTU, '12.04'): ['git'],
            (UBUNTU, '14.04'): ['git'],
        }
    
    def set_defaults(self):
        pass
        
    def get_logs_between_commits(self, a, b):
        """
        Retrieves all commit messages for all commits between the given commit numbers
        on the current branch.
        """
        ret = self.local('git log --pretty=oneline %s...%s' % (a, b), capture=True)
        if self.verbose:
            print(ret)
        return str(ret)

    def get_current_commit(self):
        """
        Retrieves the git commit number of the current head branch.
        """
        s = str(self.local('git rev-parse HEAD', capture=True))
        if self.verbose:
            print('current commit:', s)
        return s
    
    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        data = {
            'current_commit': self.get_current_commit(),
        }
        return data
    
    @task(precursors=['packager', 'pip', 'tarball', 'djangomedia', 'djangomigrations'])
    def configure(self):
        from burlap.jirahelp import update_tickets_from_git
        update_tickets_from_git()


gitchecker = GitCheckerSatchel()
gittracker = GitTrackerSatchel()



"""
Python packages
===============

This module provides tools for installing Python packages using
the ``easy_install`` command provided by `setuptools`_.

.. _setuptools: http://pythonhosted.org/setuptools/

"""

from fabric.api import cd, run

from burlap.utils import download, run_as_root


EZ_SETUP_URL = 'https://bootstrap.pypa.io/ez_setup.py'


def package_version(name, python_cmd='python'):
    """
    Get the installed version of a package

    Returns ``None`` if it can't be found.
    """
    cmd = '''%(python_cmd)s -c \
        "import pkg_resources;\
        dist = pkg_resources.get_distribution('%(name)s');\
        print dist.version"
        ''' % locals()
    res = run(cmd, quiet=True)
    if res.succeeded:
        return res
    else:
        return None


def is_setuptools_installed(python_cmd='python'):
    """
    Check if `setuptools`_ is installed.

    .. _setuptools: http://pythonhosted.org/setuptools/
    """
    version = package_version('setuptools', python_cmd=python_cmd)
    return (version is not None)


def install_setuptools(python_cmd='python', use_sudo=True):
    """
    Install the latest version of `setuptools`_.

    ::

        import burlap

        burlap.python_setuptools.install_setuptools()

    """

    setuptools_version = package_version('setuptools', python_cmd)
    distribute_version = package_version('distribute', python_cmd)

    if setuptools_version is None:
        _install_from_scratch(python_cmd, use_sudo)
    else:
        if distribute_version is None:
            _upgrade_from_setuptools(python_cmd, use_sudo)
        else:
            _upgrade_from_distribute(python_cmd, use_sudo)


def _install_from_scratch(python_cmd, use_sudo):
    """
    Install setuptools from scratch using installer
    """

    with cd("/tmp"):
        download(EZ_SETUP_URL)

        command = '%(python_cmd)s ez_setup.py' % locals()
        if use_sudo:
            run_as_root(command)
        else:
            run(command)

        run('rm -f ez_setup.py')


def _upgrade_from_setuptools(python_cmd, use_sudo):
    """
    Upgrading from setuptools 0.6 to 0.7+ is supported
    """
    _easy_install(['-U', 'setuptools'], python_cmd, use_sudo)


def _upgrade_from_distribute(python_cmd, use_sudo):
    """
    Upgrading from distribute 0.6 to setuptools 0.7+ directly is not
    supported. We need to upgrade distribute to version 0.7, which is
    a dummy package acting as a wrapper to install setuptools 0.7+.
    """
    _easy_install(['-U', 'distribute'], python_cmd, use_sudo)


def install(packages, upgrade=False, use_sudo=False, python_cmd='python'):
    """
    Install Python packages with ``easy_install``.

    Examples::

        import burlap

        # Install a single package
        burlap.python_setuptools.install('package', use_sudo=True)

        # Install a list of packages
        burlap.python_setuptools.install(['pkg1', 'pkg2'], use_sudo=True)

    .. note:: most of the time, you'll want to use
              :py:func:`burlap.python.install()` instead,
              which uses ``pip`` to install packages.

    """
    argv = []
    if upgrade:
        argv.append("-U")
    if isinstance(packages, basestring):
        argv.append(packages)
    else:
        argv.extend(packages)
    _easy_install(argv, python_cmd, use_sudo)


def _easy_install(argv, python_cmd, use_sudo):
    """
    Install packages using easy_install

    We don't know if the easy_install command in the path will be the
    right one, so we use the setuptools entry point to call the script's
    main function ourselves.
    """
    command = """python -c "\
        from pkg_resources import load_entry_point;\
        ez = load_entry_point('setuptools', 'console_scripts', 'easy_install');\
        ez(argv=%(argv)r)\
    """ % locals()
    if use_sudo:
        run_as_root(command)
    else:
        run(command)
